import io
import shutil
import zipfile
from collections import defaultdict
from pathlib import Path

import requests

# Defaults
DEFAULT_BASE_PATH = Path.home() / "Documents" / "pybo" / "dialect_packs"
DEFAULT_DIALECT_PACK = "general"


def get_dialect_pack_url(dialect_name, version=None):
    response = requests.get(
        "https://api.github.com/repos/Esukhia/botok-data/releases/latest"
    )
    if not version:
        version = response.json()["tag_name"]
    return f"https://github.com/Esukhia/botok-data/releases/download/{version}/{dialect_name}.zip"


def get_dialect_pack(dialect_name, out_dir, version=None):
    out_dir = Path(out_dir)
    out_dir.mkdir(exist_ok=True, parents=True)
    dialect_pack_path = out_dir / dialect_name
    if dialect_pack_path.is_dir():
        return dialect_pack_path

    print(f"[INFO] Downloading {dialect_name} dialect pack ...")
    # Download the dialect pack
    url = get_dialect_pack_url(dialect_name, version)
    r = requests.get(url, stream=True, timeout=50)

    # attempt 50 times to download the zip
    check = zipfile.is_zipfile(io.BytesIO(r.content))
    attempts = 0
    while not check and attempts < 50:
        r = requests.get(url, stream=True, timeout=50)
        check = zipfile.is_zipfile(io.BytesIO(r.content))
        attempts += 1

    if not check:
        raise IOError("the .zip file couldn't be downloaded.")
    else:
        # extract the zip in the current folder
        z = zipfile.ZipFile(io.BytesIO(r.content))
        z.extractall(path=str(out_dir))

    print(f"[INFO] Download completed!")

    return dialect_pack_path


class Config:
    """botok config for Tibetan dialect pack.

    Each dialect pack has two components:
      1. Dictionary:
         - contains all the data required to construct the Trie.
         - It should in the directory called `dictionary` inside the dialect pack directory.
      2. Adjustment:
         - Contains all the data required to adjust the text segmentation rules.
    """

    def __init__(self, dialect_name=None, base_path=None):
        """Create config for given `dialect_name` and stored in `base_path`"""
        if not dialect_name:
            dialect_name = DEFAULT_DIALECT_PACK
        if not base_path:
            base_path = DEFAULT_BASE_PATH
        dialect_pack_path = get_dialect_pack(dialect_name, base_path)
        self.reset(dialect_pack_path)

    def reset(self, dialect_pack_path=None):
        """Reset the config to default bo_general_pack."""
        if dialect_pack_path:
            self.dialect_pack_path = dialect_pack_path
        else:
            self.dialect_pack_path = get_dialect_pack(
                DEFAULT_DIALECT_PACK, DEFAULT_BASE_PATH
            )
        self.dictionary = self._get_pack_component("dictionary")
        self.adjustments = self._get_pack_component("adjustments")

    def _get_pack_component(self, pack_component_name, pack_component=None):
        """Return all the data_paths of the `pack_component.

        data_paths stored in python `dict` as per the directory
        structure of the pack component.
        """
        if not pack_component:
            pack_component = defaultdict(list)
        for path in (self.dialect_pack_path / pack_component_name).iterdir():
            if not path.is_dir():
                continue
            data_type = path.name
            pack_component[data_type].extend(list(path.rglob("*.tsv")))
        return pack_component

    @classmethod
    def from_path(cls, dialect_pack_path):
        """Creates config from ``dialect_pack_path``.

        Returns:
            :class: `Config`: An instance of a Configuration object

        Examples::

            config = Config.from_path(path_to_dialect_pack)
            assert config.dictionary == True
            assert config.adjustments == True

        """
        path = Path(dialect_pack_path)
        dialect_name = path.name
        base_path = path.parent
        return cls(dialect_name, base_path)

    @property
    def profile(self):
        """Returns profile name of the dialect_pack."""
        return self.dialect_pack_path.name

    def add_dialect_pack(self, path):
        """"Merge given dialect_pack at `path` to current dialect_pack."""
        self.dialect_pack_path = path
        self._get_pack_component("dictionary", self.dictionary)
        self._get_pack_component("adjustments", self.adjustments)



# coding: utf-8
from ..third_party.cqlparser import Query


class CQLMatcher:
    def __init__(self, query):
        """
        Creates a matcher object to be later executed against a list of tokens with BoMatcher.match()

        :param query: CQL compliant query string
        :type query: string

        """
        self.query = Query(query)

    def match(self, tokens_list):
        """
        Runs cql.Query on a slice of the list of tokens for every index in the list.

        :param tokens_list: output of BoTokenizer
        :type tokens_list: list of Token objects
        :return: a list of matching slices of tokens_list
        :rtype: list of tuples with each two values: beginning and end indices
        """
        slice_len = len(self.query.tokenexprs) - 1
        matches = []
        for i in range(len(tokens_list)):
            if i + slice_len <= len(tokens_list) and self.query(
                tokens_list[i : i + slice_len + 1]
            ):
                matches.append((i, i + slice_len))
        return matches



# coding: utf-8
from .chunkframework import ChunkFramework
from ..vars import ChunkMarkers as c
from ..vars import CharMarkers as a


class Chunks(ChunkFramework):
    """
    Produces chunks of the following types: bo, non-bo, punct and syl chunks

    Implements the following chunking pipeline:
            chunk "input_str" into BO / OTHER
            | chunk BO into PUNCT / BO
            | chunk BO into SYM / BO
            | chunk BO into NUM / BO
            | chunk BO into TEXT (syllables)
            | chunk OTHER into CJK / OTHER
            | chunk OTHER into LATIN / OTHER

    .. note:: Following Tibetan usage, it does not consider space as a punctuation mark.
    Spaces get attached to the chunk preceding them.
    """

    def __init__(self, string, ignore_chars=None):
        ChunkFramework.__init__(self, string, ignore_chars=ignore_chars)

    def make_chunks(self, indices=True, gen=False, space_as_punct=False):
        chunks = self.chunk_bo_chars()
        if space_as_punct:
            chunks = self.pipe_chunk(
                chunks, self.chunk_spaces, to_chunk_marker=c.BO.value, yes=c.PUNCT.value
            )
        chunks = self.pipe_chunk(
            chunks, self.chunk_punct, to_chunk_marker=c.BO.value, yes=c.PUNCT.value
        )
        chunks = self.pipe_chunk(chunks, self.chunk_symbol, c.BO.value, c.SYM.value)
        chunks = self.pipe_chunk(chunks, self.chunk_number, c.BO.value, c.NUM.value)
        if not space_as_punct:
            chunks = self.merge_skippable_punct(
                chunks
            )  # ensure we have correctly built syls
        chunks = self.pipe_chunk(chunks, self.syllabify, c.BO.value, c.TEXT.value)
        chunks = self.pipe_chunk(chunks, self.adjust_syls, c.TEXT.value, c.TEXT.value)
        chunks = self.pipe_chunk(chunks, self.chunk_cjk, c.OTHER.value, c.CJK.value)
        chunks = self.pipe_chunk(chunks, self.chunk_latin, c.OTHER.value, c.LATIN.value)
        if not space_as_punct:
            chunks = self.merge_skippable_punct(chunks)
        if not indices:
            return self.get_chunked(chunks, gen=gen)
        return chunks


class TokChunks(Chunks):
    """
    This class uses the chunks produced by ``Chunks`` to identify Tibetan syllables and clean them.
    Thus produces pre-processed Tibetan text that can be further processed.

    Every chunk produced by ``Chunks`` is wrapped into a tuple containing:
            - either None or a list containing the cleaned syllable
              (the indices to every non-space and non-tsek char in every syllable chunk)
            - the chunk itself

    """

    def __init__(self, string, ignore_chars=None, space_as_punct=False):
        super().__init__(string, ignore_chars=ignore_chars)
        self.chunks = None
        self.space_as_punct = space_as_punct

    def serve_syls_to_trie(self):
        chunks = []
        for chunk in self.make_chunks(space_as_punct=self.space_as_punct):
            if chunk[0] == c.TEXT:
                syl = self.__get_text_chars(chunk[1], chunk[1] + chunk[2])
                chunks.append((syl, chunk))
            else:
                chunks.append((None, chunk))
        self.chunks = chunks

    def get_syls(self):
        syls = []
        for chunk in self.make_chunks(space_as_punct=self.space_as_punct):
            if chunk[0] == c.TEXT:
                char_idxs = self.__get_text_chars(chunk[1], chunk[1] + chunk[2])
                syls.append("".join([self.bs.string[i] for i in char_idxs]))
        return syls

    def __get_text_chars(self, start_idx, end_idx):
        """
        Removes all the spaces and tseks from a given syllable by only keeping the characters that
        pass ``__is_syl_text()``.

        :param start_idx: starting index of the syllable-chunk to clean
        :param end_idx: its ending index
        :type start_idx: int
        :type end_idx: int
        :return: a list of indices corresponding to the chars of the cleaned syllable
        """
        return [i for i in range(start_idx, end_idx) if self.__is_syl_text(i)]

    def __is_syl_text(self, char_idx):
        """
        Tests whether the character at the given index is part of the cleaned syllable or not.
        """
        return (
            self.bs.base_structure[char_idx] != a.TSEK
            and self.bs.base_structure[char_idx] != a.TRANSPARENT
            and self.bs.base_structure[char_idx] != a.SKRT_LONG_VOW
        ) or self.bs.base_structure[char_idx] == a.SKRT_LONG_VOW



# coding: utf8
from .chunkframeworkbase import ChunkFrameworkBase
from ..vars import CharMarkers as a, ChunkMarkers as u, VOWELS, NO_SHAD_CONS


class ChunkFramework(ChunkFrameworkBase):
    """
    This class is a framework to split a given string into chunks of characters sharing similar properties.
    Fine chunking is possible by piping any number of chunking methods as desired.

    Building on top of ``BoString`` that acts as the provider of the group to which each character belongs,
    this class manipulates these groups in order to produce chunks.
    Thus, the finer the groups of characters encoded in BoString, the greater the capacities of this class
    to produce refined chunking.

    By default, this classes provides methods that should cater standard chunking needs for Tibetan language.

    :Example:

    >>> from botok.chunks.chunkframework import ChunkFramework

    >>> bo_str = ' བཀྲ་ཤིས་  tr བདེ་ལེགས།'
    >>> bc = ChunkFramework(bo_str)

    # 1. Initial chunking
    >>> chunks = bc.chunk_bo_chars()  # uses default chunk-marks here, but override them in the pipeline

    >>> chunks  # actual chunks. each chunk is tuple containing: (chunk-marker, starting_index, chunk_length)
    [(100, 0, 11), (101, 11, 2), (100, 13, 10)]
    >>> bc.get_markers(chunks)  # shows the human-readable description of the constant
    [('bo', 0, 11), ('non-bo', 11, 2), ('bo', 13, 10)]
    >>> bc.get_chunked(chunks)  # shows the substring for each chunk
    [(100, ' བཀྲ་ཤིས་  '), (101, 'tr'), (100, ' བདེ་ལེགས།')]

    # 2. First piped chunking: re-chunks tibetan chunks into tibetan text and tibetan punctuation.
    >>> bc.pipe_chunk(chunks, bc.chunk_punct, to_chunk_marker=u.BO, yes=u.PUNCT)

    >>> chunks
    [(100, 0, 11), (101, 11, 2), (100, 13, 9), (102, 22, 1)]
    >>> bc.get_markers(chunks)
    [('bo', 0, 11), ('non-bo', 11, 2), ('bo', 13, 9), ('punct', 22, 1)]
    >>> bc.get_chunked(chunks)
    [(100, ' བཀྲ་ཤིས་  '), (101, 'tr'), (100, ' བདེ་ལེགས'), (102, '།')]

    # 3. Second piped chunking: re-chunks tibetan text into syllables, keeping the same chunk-marker.
    >>> bc.pipe_chunk(chunks, bc.syllabify, to_chunk_marker=u.BO, yes=u.BO)

    >>> chunks
    [(100, 0, 5), (100, 5, 4), (100, 9, 2), (101, 11, 2), (100, 13, 5), (100, 18, 4), (102, 22, 1)]
    >>> bc.get_markers(chunks)
    [('bo', 0, 5), ('bo', 5, 4), ('bo', 9, 2), ('non-bo', 11, 2), ('bo', 13, 5), ('bo', 18, 4), ('punct', 22, 1)]
    >>> bc.get_chunked(chunks)
    [(100, ' བཀྲ་'), (100, 'ཤིས་'), (100, '  '), (101, 'tr'), (100, ' བདེ་'), (100, 'ལེགས'), (102, '།')]

    # 4. Formatting the resultant chunks into an easily usable structure
    >>> chunks = bc.get_markers(chunks)         # exchange the constants for the human-readable description
    >>> final_result = bc.get_chunked(chunks)   # exchange the indices for the substrings of each chunk
    >>> final_result
    [('bo', ' བཀྲ་'), ('bo', 'ཤིས་'), ('bo', '  '), ('non-bo', 'tr'), ('bo', ' བདེ་'), ('bo', 'ལེགས'), ('punct', '།')]



    The Framework
    =============
    This class is meant to be extended as easily as it gets by creating custom chunking components.
    A "chunking kit" comprises:
                        - a chunking method
                        - a test method
                        - marker constants (if the existing one can't be used)
                        - a human-friendly description of the chunk type in the form of
                          an entry in ``__init__().chunk_markers``(in case a new constant was created).


    Chunking Methods (exposed)
    --------------------------
    A chunking method parses the characters in the slice of the input string as described
    in its arguments("start" and "end"). It produces consecutive groups of characters that
    pass the test that is used internally and that don't pass it.

    It is expected that they all share the same syntax except for the test method it calls
    internally.


    Test methods (hidden)
    -----------------------
    A test method contains the logic behind a chunking method.
    It leverages the character groups provided by ``BoString`` to encode the semantic properties the chunks
    are expected to possess.

    For example, ``__is_bo_unicode()`` passes for every character within the Tibetan Unicode Table.
    It does so by checking that the group of the character at the given index is not OTHER.


    Piped Chunking
    --------------
    Finely chunking a string is possible by piping as many chunking methods as necessary in a specific order.

    This is done by using ``pipe_chunk()`` that takes as arguments:
            - the output of a previous chunking method,
            - a new chunking method and
            - a chunk-marker to identify on which chunks the new method should be applied.
    It will replace in place every chunk matching the chunk-marker with the results of the new chunking method.

    For example, a standard pipeline for Tibetan would be:
            chunk "input_str" into "bo"/"non-bo" | chunk "bo" into "punct"/"bo-text" | chunk "bo-text" into "syllables"

    The result would be a succession of chunks that are either tibetan syllables with their punctuation,
    non-syllabic punctuation or non-Tibetan characters.


    Custom Chunking Methods
    -----------------------
    - signature: ``chunk_xxxx(start=None, end=None, yes=<int>, no=<int>)``
                 where the ints in "yes" and "no" are the hard-coded values of the chunk's types.
    - body: ``return self.__chunk_using(self.<private_test_method>, start, end, yes, no)``
             where "private_test_method" is the custom test method where the chunking logic is encoded.


    Custom Test Methods
    -------------------
    - signature: ``__is_xxx(char_idx)``

    - body: ``return <conditions>``
            where each condition follows this pattern:
            ``self.base_structure[char_idx] <equals or not> <character group>``


    Naming conventions
    ------------------
    - chunking methods start with "chunk_xxx".
    - test methods start with "__is_xxx".
    - chunk type constants end in "_MARKER" to be differenciated from ``BoString`` constants
    """

    def chunk_bo_chars(self, start=None, end=None, yes=u.BO.value, no=u.OTHER.value):
        """
        Chunks input into Tibetan valid characters("bo") or something else("non-bo").

        :param start: starting index in ``__init__().string``
        :param end: ending index in ``__init__().string``
        :param yes: chunk-mark to apply to chunks passing the test
        :param no: chunk-mark to apply those not passing the test
        :type start: int
        :type end: int
        :type yes: int (hard-coded value of BO_MARKER)
        :type no: int (hard-coded value of NON_BO_MARKER)
        :return: the resulting chunks
        :rtype: list of tuples containing each 3 ints: chunk-mark, starting_index, chunk_length
        """
        return self.chunk_using(self.__is_bo_unicode, start, end, yes, no)

    def __is_bo_unicode(self, char_idx):
        """
        Tests whether the character at the given index is found within the Tibetan Unicode Table.

        :param char_idx: index of the character to test
        :type char_idx: int
        """
        return (
            self.bs.base_structure[char_idx] != a.OTHER
            and self.bs.base_structure[char_idx] != a.LATIN
            and self.bs.base_structure[char_idx] != a.CJK
        )

    def chunk_latin(self, start=None, end=None, yes=u.LATIN.value, no=u.OTHER.value):
        return self.chunk_using(self.__is_latin, start, end, yes, no)

    def __is_latin(self, char_idx):
        return (
            self.bs.base_structure[char_idx] == a.LATIN
            or self.bs.base_structure[char_idx] == a.TRANSPARENT
        )

    def chunk_cjk(self, start=None, end=None, yes=u.CJK.value, no=u.OTHER.value):
        return self.chunk_using(self.__is_cjk, start, end, yes, no)

    def __is_cjk(self, char_idx):
        return (
            self.bs.base_structure[char_idx] == a.CJK
            or self.bs.base_structure[char_idx] == a.TRANSPARENT
        )

    def chunk_punct(
        self, start=None, end=None, yes=u.PUNCT.value, no=u.NON_PUNCT.value
    ):
        """
        Chunks input into Tibetan text("punct") or non-Tibetan("non-punct").

        :type yes: int (hard-coded value of PUNCT_MARKER)
        :type no: int (hard-coded value of NON_PUNCT_MARKER)
        """
        return self.chunk_using(self.__is_punct, start, end, yes, no)

    def __is_punct(self, char_idx):
        """
        Tests whether the character at the given index is a Tibetan punctuation or not.
        """
        # if a tsek or a space is right after
        if (
            char_idx
            and (
                self.bs.base_structure[char_idx - 1] == a.SYMBOL
                or self.bs.base_structure[char_idx - 1] == a.NUMERAL
                or self.bs.base_structure[char_idx - 1] == a.OTHER
                or self.bs.base_structure[char_idx - 1] == a.NORMAL_PUNCT
                or self.bs.base_structure[char_idx - 1] == a.SPECIAL_PUNCT
                or self.bs.base_structure[char_idx - 1] == a.TSEK
                or self.bs.base_structure[char_idx - 1] == a.TRANSPARENT
            )
            and (
                self.bs.base_structure[char_idx] == a.TSEK
                or self.bs.base_structure[char_idx] == a.TRANSPARENT
                or self.bs.base_structure[char_idx] == a.NORMAL_PUNCT
            )
        ):
            return True

        return (
            self.bs.base_structure[char_idx] == a.NORMAL_PUNCT
            or self.bs.base_structure[char_idx] == a.SPECIAL_PUNCT
            or self.bs.base_structure[char_idx] == a.TRANSPARENT
        )

    def chunk_symbol(self, start=None, end=None, yes=u.SYM.value, no=u.NON_SYM.value):
        """
        Chunks input into Tibetan text("sym") or non-Tibetan("non-sym").

        :type yes: int (hard-coded value of SYM_MARKER)
        :type no: int (hard-coded value of NON_SYM_MARKER)
        """
        return self.chunk_using(self.__is_sym, start, end, yes, no)

    def __is_sym(self, char_idx):
        """
        Tests whether the character at the given index is a Tibetan symbols or not.
        Note: Since 0F00 (ༀ) is categorized in unicode's symbols, all the other NFC characters
                are also considered to be part of symbol chunks.
        """
        return (
            self.bs.base_structure[char_idx] == a.SYMBOL
            or self.bs.base_structure[char_idx] == a.TRANSPARENT
            or self.bs.base_structure[char_idx] == a.NFC
        )

    def chunk_number(self, start=None, end=None, yes=u.NUM.value, no=u.NON_NUM.value):
        """
        Chunks input into Tibetan text("num") or non-Tibetan("non-num").

        :type yes: int (hard-coded value of NUM_MARKER)
        :type no: int (hard-coded value of NON_NUM_MARKER)
        """
        return self.chunk_using(self.__is_num, start, end, yes, no)

    def __is_num(self, char_idx):
        """
        Tests whether the character at the given index is a number  or not.
        """
        return (
            self.bs.base_structure[char_idx] == a.NUMERAL
            or self.bs.base_structure[char_idx] == a.TRANSPARENT
        )

    def chunk_spaces(
        self, start=None, end=None, yes=u.SPACE.value, no=u.NON_SPACE.value
    ):
        """
        Chunks input into any valid Unicode spaces("space") or something else("non-space").

        :type yes: int (hard-coded value of SPACE_MARKER)
        :type no: int (hard-coded value of NON_SPACE_MARKER)
        """
        return self.chunk_using(self.__is_space, start, end, yes, no)

    def __is_space(self, char_idx):
        """
        Tests whether the character at the given index is a valid Unicode space or not.
        """
        return self.bs.base_structure[char_idx] == a.TRANSPARENT

    def syllabify(self, start=None, end=None, yes=u.TEXT.value):
        """
        Chunks valid Tibetan text(expected input) into syllables(tsek is included if present).

        :type yes: int (hard-coded value of SYL_MARKER)
        """
        if not start and not end:
            start, end = 0, self.bs.len

        indices = self.chunk(start, end, self.__is_tsek_or_long_skrt_vowel)
        for num, i in enumerate(indices):
            if i[0] and num - 1 >= 0 and not indices[num - 1][0]:
                indices[num - 1] = (
                    indices[num - 1][0],
                    indices[num - 1][1],
                    indices[num - 1][2] + i[2],
                )

        return [(yes, i[1], i[2]) for i in indices if not i[0]]

    def __is_tsek_or_long_skrt_vowel(self, char_idx):
        """
        Tests whether the character at the given index is an unambiguous end-marker or not.
        Used as test to find syllable boundaries by ``syllabify()``.
        ::note:
        """
        return (
            self.bs.base_structure[char_idx] == a.TSEK
            or self.bs.base_structure[char_idx] == a.SKRT_LONG_VOW
        )

    def adjust_syls(self, start=None, end=None, yes=u.TEXT.value):
        """
        if there is a space in the chunk:
                test if the preceding is ཀ ག ཤ eventually followed by a vowel
                test if both parts are valid syllables (Bosyl)
                if tests pass:
                        split in two syllables after space.
        :param chunks:
        :param start:
        :param end:
        :param yes:
        :return:
        """
        indices = self.chunk(start, end, self.__is_transparent)
        for num, i in enumerate(indices):
            if len(indices) - 1 > num > 0 and indices[num][0]:
                _, s, e = indices[num - 1]
                text = self.bs.string[s : s + e]
                if (
                    len(text) >= 2 and text[-1] in VOWELS and text[-2] in NO_SHAD_CONS
                ) or (len(text) >= 1 and text[-1] in NO_SHAD_CONS):
                    indices[num - 1] = (
                        yes,
                        indices[num - 1][1],
                        indices[num - 1][2] + i[2],
                    )
                else:
                    indices[num - 1] = (
                        indices[num - 1][0],
                        indices[num - 1][1],
                        indices[num - 1][2] + indices[num][2] + indices[num + 1][2],
                    )
                    indices[num + 1] = (None, indices[num + 1][1], indices[num + 1][2])
            elif indices[num][0] is False:
                indices[num] = (yes, i[1], i[2])
            elif (num == 0 or num == len(indices) - 1) and indices[num][0] is True:
                indices[num] = (u.PUNCT.value, i[1], i[2])

        # remove all chunks that were added
        indices = [i for i in indices if i[0] is not True and i[0] is not None]

        return indices if len(indices) > 1 else list()

    def __is_transparent(self, char_idx):
        return self.bs.base_structure[char_idx] == a.TRANSPARENT



import pytest

from botok import Config, Tokenize, WordTokenizer


@pytest.fixture(scope="session")
def empty_wt():
    """Return empty word tokenizer."""
    config = Config.from_path("./tests/data/empty_dialect_pack")
    return WordTokenizer(config=config)


@pytest.fixture(scope="session")
def wt():
    """Return default word tokenizer."""
    return WordTokenizer()



# coding: utf-8
from .token import Token
from ..vars import NAMCHE, TSEK
from ..vars import chunk_values as u
from ..vars import char_values as a
from ..vars import CharMarkers as A
from ..vars import WordMarkers as w
from ..third_party.has_skrt_syl import has_skrt_syl


class Tokenize:
    """
    Expects a BoTrie instance as trie
    """

    def __init__(self, trie):
        self.pre_processed = None
        self.trie = trie

    def tokenize(self, pre_processed, debug=False):
        """

        :param pre_processed: PyBoTextChunks of the text to be tokenized
        :param debug: prints debug info in True
        :return: a list of Token objects
        """
        self.pre_processed = pre_processed
        tokens = []

        c_idx = 0
        while c_idx < len(self.pre_processed.chunks):
            walker = c_idx
            syls = []
            max_match = []
            match_data = {}
            current_node = None
            found_max_match = False

            while True:
                cur_syl = self.pre_processed.chunks[walker][0]
                # CHUNK IS SYLLABLE
                if cur_syl:
                    syl = ''.join([self.pre_processed.bs.string[i] for i in cur_syl])
                    current_node = self.trie.walk(syl, current_node)
                    if current_node:
                        syls.append(walker)
                        if current_node.is_match():
                            match_data[walker] = current_node.data
                            max_match.append(syls[:])
                            # check if the matched is last
                            if walker + 1 == len(self.pre_processed.chunks):
                                found_max_match = True
                        else:
                            if walker + 1 == len(self.pre_processed.chunks):
                                if max_match:
                                    found_max_match = True
                                else:
                                    # OOV syllables are turned into independant tokens
                                    self.add_found_word_or_non_word(
                                        walker, match_data, syls, tokens
                                    )
                                    c_idx += len(syls)
                                    break
                    # CAN'T CONTINUE WALKING
                    else:
                        # max_match followed by syllable
                        if max_match:
                            found_max_match = True
                        else:
                            # check if syllables is NO_POS or Non-word
                            if syls:
                                self.add_found_word_or_non_word(
                                    walker, match_data, syls, tokens
                                )
                                c_idx += len(syls)
                                break
                            # syllabel is not in the dictionary (Trie)
                            else:
                                non_word = [walker]
                                tokens.append(
                                    self.chunks_to_token(non_word, {}, ttype=w.NON_WORD.name)
                                )
                                c_idx += 1
                                break
                # CHUNK IS NON-SYLLABLE
                else:
                    # max_match followed by non-syllable
                    if max_match:
                        found_max_match = True
                    # check for any syllables left, which are to be turned into independant tokens
                    elif syls:
                        self.add_found_word_or_non_word(
                            walker, match_data, syls, tokens
                        )
                        c_idx += len(syls)
                        break
                    # non-syllable are turned into independant tokens
                    else:
                        tokens.append(
                            self.chunks_to_token([c_idx], {})
                        )
                        c_idx += 1
                        break

                if found_max_match:
                    self.add_found_word_or_non_word(
                        c_idx+len(max_match[-1])-1, match_data, max_match[-1], tokens
                    )
                    c_idx += len(max_match[-1])
                    break

                walker += 1

        self.pre_processed = None

        return tokens

    def add_found_word_or_non_word(
        self, c_idx, match_data, syls, tokens, has_decremented=False
    ):
        # there is a match
        if c_idx in match_data.keys():
            data = match_data[c_idx]
            ttype = (
                w.NO_POS.name
                if "senses" not in data
                or len([True for m in data["senses"] if "pos" in m]) <= 0
                else None
            )
            tokens.append(self.chunks_to_token(syls, data, ttype=ttype))
        elif any(match_data.values()):
            non_max_idx = sorted(match_data.keys())[-1]
            non_max_syls = []
            for syl in syls:
                if syl <= non_max_idx:
                    non_max_syls.append(syl)
            data = match_data[non_max_idx]
            ttype = (
                w.NO_POS.name
                if "senses" not in data
                or len([True for m in data["senses"] if "pos" in m]) <= 0
                else None
            )
            tokens.append(self.chunks_to_token(non_max_syls, data, ttype=ttype))
            c_idx = non_max_idx
        else:
            # add first syl in syls as non-word
            tokens.append(self.chunks_to_token([syls[0]], {}, ttype=w.NO_POS.name))

            # decrement chunk-idx for a new attempt to find a match
            if syls:
                c_idx -= len(syls[1:]) - 1
            if (
                has_decremented
                or (
                    c_idx < len(self.pre_processed.chunks)
                    and self.pre_processed.chunks[c_idx][0] is None
                )
                or len(syls) > 1
            ):
                c_idx -= 1
        return c_idx

    def chunks_to_token(self, syls, data, ttype=None):
        if len(syls) == 1:
            # chunk format: ([char_idx1, char_idx2, ...], (type, start_idx, len_idx))
            token_syls = [self.pre_processed.chunks[syls[0]][0]]
            token_type = self.pre_processed.chunks[syls[0]][1][0]
            token_start = self.pre_processed.chunks[syls[0]][1][1]
            token_length = self.pre_processed.chunks[syls[0]][1][2]
            syl_start_end = [
                (
                    self.pre_processed.chunks[syls[0]][1][1],
                    self.pre_processed.chunks[syls[0]][1][2],
                )
            ]
            if ttype:
                if "senses" not in data:
                    data["senses"] = [{"pos": ttype}]
                else:
                    for m in data["senses"]:
                        if "pos" not in m:
                            m["pos"] = ttype

            return self.create_token(
                token_type, token_start, token_length, token_syls, syl_start_end, data
            )
        elif len(syls) > 1:
            token_syls = [self.pre_processed.chunks[idx][0] for idx in syls]
            token_type = self.pre_processed.chunks[syls[-1]][1][0]
            token_start = self.pre_processed.chunks[syls[0]][1][1]
            token_length = 0
            syl_start_end = []
            for i in syls:
                token_length += self.pre_processed.chunks[i][1][2]
                syl_start_end.append(
                    (
                        self.pre_processed.chunks[i][1][1],
                        self.pre_processed.chunks[i][1][2],
                    )
                )
            if ttype:
                if "senses" not in data:
                    data["senses"] = [{"pos": ttype}]
                else:
                    for m in data["senses"]:
                        if "pos" not in m:
                            m["pos"] = ttype

            return self.create_token(
                token_type, token_start, token_length, token_syls, syl_start_end, data
            )
        else:
            raise ValueError(str(syls) + "should contain at least 1 token")

    def create_token(self, ttype, start, length, syls, syl_start_end, data):
        """
        :param ttype: token type
        :param start: start index in input string
        :param length: length of the substring from the input string corresponding to this token
        :param syls: syl representation coming from PyBoTextChunks.
                        the indices are modified to be usable on the substring corresponding to this token
        :param tag: the POS retrieved from the chunk or from the trie
        :param freq: the frequency retrieved from from the trie
        :return: a Token object with all the above information
        """
        token = Token()
        token.text = self.pre_processed.bs.string[start : start + length]
        token.chunk_type = u[ttype]
        token.start = start
        token.len = length
        if syls != [None]:
            token.syls_idx = [[s - start for s in syl] for syl in syls]
            token.syls_start_end = [
                {"start": s - start, "end": s - start + l} for s, l in syl_start_end
            ]
        char_groups = self.pre_processed.bs.export_groups(
            start, length, for_substring=True
        )
        token.char_types = [a[char_groups[idx]] for idx in sorted(char_groups.keys())]
        for k, v in data.items():
            token[k] = v
        token.skrt = (
            self.is_sanskrit(char_groups, token.text) if not token.skrt else token.skrt
        )
        return token

    def is_sanskrit(self, char_groups, word):
        return self._has_skrt_char(char_groups) or has_skrt_syl(word)

    def _has_skrt_char(self, char_groups):
        return (
            A.SKRT_VOW in char_groups.values()
            or A.SKRT_CONS in char_groups.values()
            or A.SKRT_SUB_CONS in char_groups.values()
        )

    @staticmethod
    def debug(debug, to_print):
        if debug:
            print(to_print, flush=True)



from pathlib import Path
import csv
from collections import defaultdict

from ..vars import CharMarkers as c

# Get the categories of Tibetan characters from the csv file
categories = defaultdict(list)
table_path = Path(__file__).parent.parent / "resources/bo_uni_table.csv"
for row in list(csv.reader(table_path.open(encoding="utf-8-sig")))[1:]:
    char = row[1].replace("—", "")
    cat = c[row[2]].value
    categories[cat].append(char)

# all unicode chars liable to be used as spaces or that allowed in a valid Tibetan string
# yet that will be ignored when read by a human.
transparent = [
    " ",  # \U32 SPACE
    "᠎",  # \U6158 MONGOLIAN VOWEL SEPARATOR
    " ",  # \U8192 EN QUAD
    " ",  # \U8193 EM QUAD
    " ",  # \U8194 EN SPACE
    " ",  # \U8195 EM SPACE
    " ",  # \U8196 THREE-PER-EM SPACE
    " ",  # \U8197 FOUR-PER-EM SPACE
    " ",  # \U8198 SIX-PER-EM SPACE
    " ",  # \U8199 FIGURE SPACE
    " ",  # \U8200 PUNCTUATION SPACE
    " ",  # \U8201 THIN SPACE
    " ",  # \U8202 HAIR SPACE
    "​",  # \U8203 ZERO WIDTH SPACE
    " ",  # \U8239 NARROW NO-BREAK SPACE
    " ",  # \U8287 MEDIUM MATHEMATICAL SPACE
    "　",  # \U12288 IDEOGRAPHIC SPACE
    "﻿",  # \U65279 ZERO WIDTH NO-BREAK SPACE
    "\t",  # Tabulation
    "\n",  # carriage return can happen in the middle of a word
]


def get_char_category(char):
    # source for codepoints: https://jrgraphix.net/research/unicode.php
    if char in transparent:
        return c.TRANSPARENT.value

    # Tibetan range
    if "\u0f00" <= char <= "\u0fff":
        for cat, chars in categories.items():
            if char in chars:
                return cat
        raise ValueError(
            f'The char "{char}" is expected to be in the tibetan table, but is not.'
        )

    # CJK range
    elif (
        "\u2e80" <= char <= "\ufaff"
        or "\ufe30" <= char <= "\ufe4f"
        or eval('"\u20000"') <= char <= eval('"\u2fa1f"')
    ):
        return c.CJK.value

    # LATIN range
    # 1. 0020 - 036f:  Latin Basic + Latin-1 Supplement + Latin Extended-A + Latin Extended-B
    # IPA Extensions + Spacing Modifier Letters + Combining Diacritical Marks
    # 2. 1e00 - 20cf: Latin Extended Additional + Superscripts and Subscripts + Currency Symbols
    elif "\u0020" <= char <= "\u036f" or "\u1e00" <= char <= "\u20cf":
        return c.LATIN.value

    else:
        return c.OTHER.value



# coding: utf8
import copy
from pathlib import Path

import pytest

from botok import Config
from botok.config import DEFAULT_BASE_PATH


@pytest.fixture(scope="module")
def base_path():
    return DEFAULT_BASE_PATH


def test_defaults(base_path):
    config = Config()

    # default dialect pach path
    assert config.dialect_pack_path == base_path / "general"
    assert config.dialect_pack_path.is_dir()

    # Trie data should be .tsv file
    for data_type in ["words", "rules"]:
        assert data_type in config.dictionary
        for data_fn in config.dictionary[data_type]:
            assert data_fn.suffix == ".tsv"

    # Segmentation adjustment
    for data_type in ["remove", "rules", "words", "words_skrt"]:
        assert data_type in config.adjustments
        for data_fn in config.adjustments[data_type]:
            if data_fn.suffix:
                assert data_fn.suffix == ".tsv"


def test_custome_dialect_pack(base_path):
    config = Config(dialect_name="kangyur")
    assert config.dialect_pack_path == base_path / "kangyur"
    assert config.dialect_pack_path.is_dir()


def test_reset(base_path):
    custome_pack_name = "kangyur"
    config = Config(dialect_name=custome_pack_name)
    assert config.dialect_pack_path == base_path / custome_pack_name

    config.reset()

    assert config.dialect_pack_path == base_path / "general"


def test_empty_config():
    config = Config.from_path("./tests/data/empty_dialect_pack")

    assert not config.dictionary
    assert not config.adjustments


def test_add_dialect_pack():
    config = Config()
    old_dictionary = copy.deepcopy(config.dictionary)
    old_adjustments = copy.deepcopy(config.adjustments)

    config.add_dialect_pack(Path("./tests/data/trie_dialect_pack"))

    assert config.dictionary != old_dictionary
    assert config.adjustments != old_adjustments



# coding: utf-8
from pathlib import Path

from .chunks.chunkframework import ChunkFramework
from .chunks.chunkframeworkbase import ChunkFrameworkBase
from .chunks.chunks import Chunks, TokChunks
from .config import Config
from .modifytokens.adjusttokens import AdjustTokens
from .modifytokens.cqlmatcher import CQLMatcher
from .modifytokens.mergedagdra import MergeDagdra
from .modifytokens.mergingmatcher import MergingMatcher
from .modifytokens.replacingmatcher import ReplacingMatcher
from .modifytokens.splitaffixed import split_affixed
from .modifytokens.splittingmatcher import SplittingMatcher
from .modifytokens.tokenmerge import TokenMerge
from .modifytokens.tokensplit import TokenSplit
from .text.pipelinebase import PipelineBase
from .text.text import Text
from .textunits.bostring import BoString
from .textunits.bosyl import BoSyl
from .textunits.sylcomponents import SylComponents
from .third_party.cqlparser import Query, parse_cql_query, replace_token_attributes
from .tokenizers.chunktokenizer import ChunkTokenizer
from .tokenizers.paragraphtokenizer import paragraph_tokenizer
from .tokenizers.sentencetokenizer import sentence_tokenizer
from .tokenizers.token import Token
from .tokenizers.tokenize import Tokenize
from .tokenizers.wordtokenizer import WordTokenizer
from .tries.basictrie import BasicTrie
from .tries.trie import Trie
from .utils.expose_data import expose_data
from .vars import *

# from .utils.get_data import get_data
from .vars import __version__



# coding: utf-8
from warnings import warn

from .charcategories import get_char_category
from ..vars import CharMarkers as a
from ..vars import char_values


class BoString:
    """
    This class is the foundational building block of pre-processing.

    It implements the natural groups of characters a user makes when looking at
    a text in his native language.

    Implementation:
    ---------------

        - all the characters in the Unicode Tables for Tibetan are organized in lists
            hard-coded as string variables in ``__attribute_basic_types()``.
        - upon instanciation, ``__init__().base_structure`` is populated with the indices of every
            char in the input string(key) and the group constant to which it belongs(values)
        - human-readable description of the group constant can be accessed in ``__init__().char_markers``

    :Example:

    >>> from botok.textunits.bostring import BoString
    >>> from botok.vars import CharMarkers

    >>> bo_str = ' བཀྲ་ཤིས་  tr བདེ་ལེགས།'
    >>> bs = BoString(bo_str)

    >>> bs.base_structure  # key: character index, value: character group
    {0: 15, 1: 1, 2: 1, 3: 2, 4: 4, 5: 1, 6: 3, 7: 1, 8: 4, 9: 15, 10: 15, 11: 14,
    12: 14, 13: 15, 14: 1, 15: 1, 16: 3, 17: 4, 18: 1, 19: 3, 20: 1, 21: 1, 22: 8}

    >>> bs.get_categories()
    {0: 'space', 1: 'cons', 2: 'cons', 3: 'sub-cons', 4: 'tsek', 5: 'cons', 6: 'vow',
    7: 'cons', 8: 'tsek', 9: 'space', 10: 'space', 11: 'other', 12: 'other',
    13: 'space', 14: 'cons', 15: 'cons', 16: 'vow', 17: 'tsek', 18: 'cons', 19: 'vow',
    20: 'cons', 21: 'cons', 22: 'punct'}

    .. note:: You may want to refine the groups that are implemented to have a finer analysis.
                Be sure to create the corresponding constants in ``__init__()`` and the corresponding
                entries in ``__init__().char_markers``.
    """

    def __init__(self, string, ignore_chars=None):
        if ignore_chars is None:
            ignore_chars = []
        self.ignore_chars = ignore_chars
        self.string = string
        self.len = len(string)
        self.base_structure = {}
        self.__attribute_basic_types()

    def __attribute_basic_types(self):
        """
        Populates ``__init__().base_structure``.
        """
        for i in range(len(self.string)):
            char = self.string[i]
            cat = get_char_category(char)
            self.__nfc_check(cat, i)
            if char in self.ignore_chars:
                self.base_structure[
                    i
                ] = (
                    a.TRANSPARENT.value
                )  # spaces chars are allowed anywhere, thus ignored
            else:
                self.base_structure[i] = cat

    def __nfc_check(self, cat, idx):
        if cat == a.NFC:
            slice_start = 10
            slice_end = 10
            while idx - slice_start < 0:
                slice_start -= 1
            while idx + slice_end >= self.len:
                slice_end -= 1
            warn(
                f'Beware of unexpected results: input string contains the non-expanded char "{self.string[idx]}", '
                f'found in "{self.string[slice_start: slice_end]}".'
            )

    def export_groups(self, start_idx, slice_len, for_substring=True):
        """
        Export the base groups for a slice of the input string

        :param start_idx: starting index of the slice
        :param slice_len: length of the slice we want to export
        :param for_substring: if True, indices start at 0, Else the indices of the original string are kept.
        :type start_idx: int
        :type slice_len: int
        :return: the slice of ``__init__().base_structure`` described in the parameters
        :rtype: dict

        :Example:

        >>> bo_str = ' བཀྲ་ཤིས་  tr བདེ་ལེགས།'
        >>> bs = BoString(bo_str)

        >>> bs.export_groups(2, 5)
        {0: 1, 1: 2, 2: 4, 3: 1, 4: 3}

        >>> bs.export_groups(2, 5, for_substring=False)
        {2: 1, 3: 2, 4: 4, 5: 1, 6: 3}

        """
        if for_substring:
            return {
                n: self.base_structure[i]
                for n, i in enumerate(range(start_idx, start_idx + slice_len))
            }
        else:
            return {
                i: self.base_structure[i]
                for i in range(start_idx, start_idx + slice_len)
            }

    def get_categories(self, struct=None):
        if struct is None or not isinstance(struct, dict):
            return {k: char_values[v] for k, v in self.base_structure.items()}
        else:
            return {k: char_values[v] for k, v in struct.items()}



# coding: utf-8
from ..textunits.bostring import BoString
from ..vars import chunk_values
from ..vars import CharMarkers as a
from ..vars import ChunkMarkers as u


class ChunkFrameworkBase:
    def __init__(self, string, ignore_chars=None):
        self.bs = BoString(string, ignore_chars=ignore_chars)

    @staticmethod
    def merge_chunks(chunks, merge_condition_func):
        """

        :param chunks: full list of chunks
        :param merge_condition_func: function that takes start-end indices of the sub-string of both
                                        previous and current chunks, does a test and returns a bool
        """
        num = 0
        while num <= len(chunks) - 1:
            current = chunks[num]
            if num - 1 >= 0:
                previous = chunks[num - 1]
                if merge_condition_func(previous, current):
                    inc = 0
                    while not previous:
                        inc += 1
                        previous = chunks[num - 1 - inc]

                    new = (previous[0], previous[1], previous[2] + current[2])
                    chunks[num - 1 - inc] = (
                        previous[0],
                        previous[1],
                        previous[2] + current[2],
                    )
                    del chunks[num]
                    num -= 1

            num += 1
        return chunks

    @staticmethod
    def merge_condition(chunk, condition_func):
        start, end = chunk[1], chunk[1] + chunk[2]
        count = 0
        i = start
        while i < end:
            if condition_func(i):
                count += 1
            i += 1
        return count == end - start

    @staticmethod
    def pipe_chunk(chunks, piped_chunk_func, to_chunk_marker: int, yes: int):
        """
        Re-chunks in place the chunks produced by a previous chunking method.

        :param chunks: the chunks from a previous chunking method
        :param piped_chunk_func: new chunking method to apply
        :param to_chunk_marker: chunk-mark to identify which chunks will be re-chunked
        :param yes: new chunk-mark to be used for matching chunks (leave empty to use default value)
                    The new chunks not passing the internal test will keep the previous chunk-mark.
        :type chunks: list of tuples containing each 3 ints
        :type piped_chunk_func: callable
        :type yes: int
        """
        for i, chunk in enumerate(chunks):
            if chunk[0] == to_chunk_marker:
                new = piped_chunk_func(chunk[1], chunk[1] + chunk[2], yes=yes)
                if new:
                    del chunks[i]
                    for j, n_chunk in enumerate(new):
                        if n_chunk[0] != yes:
                            chunks.insert(i + j, (chunk[0], n_chunk[1], n_chunk[2]))
                        else:
                            chunks.insert(i + j, n_chunk)
        return chunks

    def chunk_using(self, condition_func, start, end, yes, no):
        if not start and not end:
            start, end = 0, self.bs.len

        indices = self.chunk(start, end, condition_func)
        return [(yes, i[1], i[2]) if i[0] else (no, i[1], i[2]) for i in indices]

    @staticmethod
    def chunk(start_idx, end_idx, condition_func):
        """
        The method that actually creates groups of characters satisfying the test method
        and not satisfying it from the given range within the input string.

        :param start_idx: first char of the range to be chunked
        :param end_idx: last char
        :param condition_func: test method
        :type start_idx: int
        :type end_idx: int
        :type condition_func: callable
        :return: the chunk indices with True/False instead of the chunk markers
        :rtype: list of tuples containing each: a bool (matched/not the test method) and the indices
        """
        chunked = []
        start = start_idx
        length = 0
        prev_state = -1
        current_state = -1
        for i in range(start_idx, end_idx):
            current_state = condition_func(i)
            if prev_state == -1:
                prev_state = current_state

            if current_state == prev_state:
                length += 1
            else:
                chunked.append((prev_state, start, length))
                prev_state = current_state
                start += length
                length = 1
        # final element
        if length != 0:
            if current_state == prev_state:
                if start + length < end_idx:
                    length += 1
            chunked.append((prev_state, start, length))
        return chunked

    def clean_chunks(self, chunks):
        chunks = self.merge_spaces(chunks)
        return self.merge_similar_chunks(chunks)

    def merge_skippable_punct(self, chunks):
        i = 0
        while i <= len(chunks) - 1:
            current = chunks[i]
            # first element
            if i == 0 and len(chunks) - 1 >= 1:
                to_del = True
                for char_idx in range(current[1], current[1] + current[2]):
                    if not self.__is_skippable_punct(char_idx):
                        to_del = False

                if to_del:
                    new_chunk = (
                        chunks[i + 1][0],
                        current[1],
                        chunks[i + 1][1] + chunks[i + 1][2],
                    )
                    chunks[i + 1] = new_chunk
                    del chunks[i]
                    i -= 1

            # remaining ones
            if i - 1 >= 0:
                to_del = True
                for char_idx in range(current[1], current[1] + current[2]):
                    if not self.__is_skippable_punct(char_idx):
                        to_del = False

                if to_del:
                    new_chunk = (
                        chunks[i - 1][0],
                        chunks[i - 1][1],
                        current[2] + chunks[i - 1][2],
                    )
                    chunks[i - 1] = new_chunk
                    del chunks[i]
                    i -= 1
            i += 1
        return self.merge_similar_chunks(chunks)

    def __is_skippable_punct(self, char_idx):
        return self.bs.base_structure[char_idx] == a.TSEK or self.__is_space(char_idx)

    def merge_spaces(self, chunks):
        return self.merge_chunks(chunks, self.__is_space_mergeable)

    def __is_space_mergeable(self, previous, current):
        # previous needs to be there as a placeholder
        return self.merge_condition(current, self.__is_space)

    def __is_space(self, char_idx):
        """
        Tests whether the character at the given index is a valid Unicode space or not.
        """
        return self.bs.base_structure[char_idx] == a.TRANSPARENT

    def merge_similar_chunks(self, chunks):
        return self.merge_chunks(chunks, self.__is_similar_chunks)

    @staticmethod
    def __is_similar_chunks(previous, current):
        # the chunk markers of both are similar.
        return (
            previous[0] != u.TEXT.value
            and current[0] != u.TEXT.value
            and previous[0] == current[0]
        )

    def get_readable(self, indices, gen=False):
        out = self.get_markers(indices)
        return self.get_chunked(out, gen=gen)

    def get_chunked(self, indices, gen=False):
        """
        Replaces the indices in every chunk tuple with the corresponding substring.

        :param indices: the output of a previous chunking method
        :param gen: if True, returns a generator with the chunks, a list otherwise
        :type indices: list of tuples each containing 3 ints
        :type gen: bool
        :return: the marker/substring pairs in a list
        :rtype: list of tuples each containing: a chunk-mark(int or str), a substring
        """
        if gen:
            return (
                (i, self.bs.string[start : start + length])
                for i, start, length in indices
            )
        return [
            (i, self.bs.string[start : start + length]) for i, start, length in indices
        ]

    @staticmethod
    def get_markers(indices):
        """
        Replaces the int representation of the chunk-mark by its str counterpart.

        :param indices: indices containing ints as markers
        :type indices: list of tuples containing each an int and the indices or the substring
        :return: the chunks where the chunk-mark is the human-readable description
        :rtype: list of tuples containing each a str and the indices or the substring
        """
        return [tuple([chunk_values[i[0]]] + list(i[1:])) for i in indices]



# coding: utf-8
from .tokensplit import TokenSplit


def split_affixed(tokens):
    """
    Splits in place the tokens containing affixed particles
    tokens have to be Token objects produced by BoTokenizer.Tokenizer

    :param tokens: list of Token objects
    """
    t = 0
    while t <= len(tokens) - 1:
        # check that splitting is possible (affixation attribute exists)
        # and that there is no meaning that has "affixed: False".
        # ie, check that the inflected form can't be the affixed form of a word and the unaffixed form of another word
        if tokens[t].affixation and not [
            True for m in tokens[t].senses if "affixed" in m and not m["affixed"]
        ]:
            # split token containing the affixed particle
            split_idx = tokens[t].syls_idx[-1][-tokens[t].affixation["len"]]
            changes = (
                '[affix_host="True"] '
                '[pos="PART" & affix="True" & skrt="False" & freq="None" & senses="None"]'
            )
            ts = TokenSplit(tokens[t], split_idx, token_changes=changes)
            token1, token2 = ts.split()
            if token2.senses is None:
                token2.senses = []

            # replace the original token with the two new ones
            tokens[t : t + 1] = [token1, token2]

            t += 1  # increment once more to account for the newly split token
        t += 1



# coding: utf-8
from enum import Enum, IntEnum

__version__ = "0.8.2"

NO_POS = "NOPOS"
TSEK = "་"
NAMCHE = "ཿ"
SHAD = "།"
AA = "འ"
HASH = "#"
VOWELS = ["ི"]
NO_SHAD_CONS = ["ཀ", "ག", "ཤ"]
DAGDRA = ["པ་", "པོ་", "བ་", "བོ་"]

CharMarkers = IntEnum(
    "CharMarkers",
    [
        # regular Tibetan
        "CONS",
        "SUB_CONS",
        "VOW",
        "TSEK",
        # punctuation
        "NORMAL_PUNCT",
        "SPECIAL_PUNCT",
        # others
        "NUMERAL",
        "SYMBOL",
        "IN_SYL_MARK",
        "NON_BO_NON_SKRT",
        # lexica_skrt
        "SKRT_CONS",
        "SKRT_SUB_CONS",
        "SKRT_VOW",
        "SKRT_LONG_VOW",
        # other languages
        "CJK",
        "LATIN",
        # misc
        "OTHER",
        "TRANSPARENT",
        "NFC",
    ],
    start=1,
)
char_values = {c.value: c.name for c in CharMarkers}

ChunkMarkers = IntEnum(
    "ChunkMarkers",
    [
        # languages
        "BO",
        "LATIN",
        "CJK",
        "OTHER",
        # tibetan textual content
        "TEXT",
        # tibetan non-textual content
        "PUNCT",
        "NON_PUNCT",
        "SPACE",
        "NON_SPACE",
        "SYM",
        "NON_SYM",
        "NUM",
        "NON_NUM",
    ],
    start=100,
)
chunk_values = {c.value: c.name for c in ChunkMarkers}

WordMarkers = IntEnum("WordMarkers", ["WORD", "NO_POS", "NON_WORD"], start=1000)
word_values = {w.value: w.name for w in WordMarkers}

Ids = Enum("Ids", ["profile", "prep", "tok", "mod", "form"])



# coding: utf8
import csv
from pathlib import Path

from ..chunks.chunks import TokChunks
from ..config import Config
from ..modifytokens.adjusttokens import AdjustTokens
from ..modifytokens.mergedagdra import MergeDagdra
from ..modifytokens.splitaffixed import split_affixed
from ..textunits.bosyl import BoSyl
from ..tries.trie import Trie
from ..vars import AA, TSEK
from .tokenize import Tokenize


def get_part_lemmas(path):
    part_lemmas = {}
    if not path.is_file():
        return part_lemmas
    with path.open("r", encoding="utf-8-sig") as f:
        reader = csv.reader(f, delimiter="\t")
        for row in list(reader)[1:]:
            form, _, lemma, _, _ = row
            part_lemmas[form] = lemma
    return part_lemmas


class WordTokenizer:
    """
    Convenience class to tokenize a given string.

    """

    def __init__(
        self, config=None, ignore_chars=None, build_trie=False,
    ):
        """
        :param tok_profile: profile for building the trie. (see config.yaml)
        """
        if not config:
            # if config is not given then use default config
            config = Config()

        self.config = config
        self.ignore_chars = ignore_chars
        self.tok = Tokenize(
            Trie(
                BoSyl,
                config.profile,
                main_data=config.dictionary,
                custom_data=config.adjustments,
                pickle_path=config.dialect_pack_path.parent,
                build=build_trie,
            )
        )

        self.adj = AdjustTokens(
            main=config.dictionary["rules"], custom=config.adjustments["rules"]
        )

        self.part_lemmas = get_part_lemmas(
            config.dialect_pack_path
            / "dictionary"
            / "words_non_inflected"
            / "particles.tsv"
        )

    def tokenize(self, string, split_affixes=True, spaces_as_punct=False, debug=False):
        """
        :param string: to be tokenized
        :param split_affixes: separates the affixed particles into seperate tokens if True
        :param debug: print debug info while parsing
        :return: list of pybo.tokenizers.Token objects
        """
        preprocessed = TokChunks(
            string, ignore_chars=self.ignore_chars, space_as_punct=spaces_as_punct
        )
        preprocessed.serve_syls_to_trie()
        tokens = self.tok.tokenize(preprocessed, debug=debug)

        if split_affixes:
            split_affixed(tokens)

        self._get_default_lemma(tokens)
        self._choose_default_entry(tokens)

        # merge pa/po/ba/bo tokens with previous ones
        MergeDagdra().merge(tokens)

        # do adjustments
        tokens = self.adj.adjust(tokens)

        return tokens

    def _get_default_lemma(self, token_list):
        for t in token_list:
            # pass any token that is not a word
            if not t.text_unaffixed:
                continue

            # otherwise, check whether the aa needs to be added and if a tsek should be added
            if t.affix and not t.affix_host:
                part = "".join(["".join(syl) for syl in t.syls])
                lemma = self.part_lemmas[part] if part in self.part_lemmas else part
                lemma += TSEK
            elif not t.affix and t.affix_host:
                lemma = (
                    t.text_unaffixed + AA + TSEK
                    if t.affixation["aa"]
                    else t.text_unaffixed + TSEK
                )
            else:
                lemma = (
                    t.text_unaffixed
                    if t.text_unaffixed.endswith(TSEK)
                    else t.text_unaffixed + TSEK
                )

            for m in t.senses:
                if "lemma" not in m and ("pos" in m and m["pos"] != "NON_WORD"):
                    m["lemma"] = lemma
            if not t.senses:
                t.senses.append({"lemma": lemma})

    @staticmethod
    def _choose_default_entry(token_list):
        def choose_n_apply(senses, t):
            s = sorted(senses, key=len, reverse=True)
            for a in ["pos", "lemma", "freq", "sense"]:
                if a in s[0]:
                    t[a] = s[0][a]

        for t in token_list:
            if t.senses:
                # Categorize all meanings in three groups
                affixed, non_affixed, no = [], [], []
                for m in t.senses:
                    if "affixed" in m:
                        if m["affixed"]:
                            affixed.append(m)
                        else:
                            non_affixed.append(m)
                    else:
                        no.append(m)

                # Decide what meaning to use as default
                # get a meaning from either group in the following order: non_affixed, no, affixed
                # take the one with the highest amount of attrs
                if non_affixed:
                    choose_n_apply(non_affixed, t)
                elif no:
                    choose_n_apply(no, t)
                elif affixed:
                    choose_n_apply(affixed, t)
                else:
                    raise ValueError("This should never happen.")



# coding: utf8
import sys

import pytest

from botok import TokChunks, Config, Trie, BoSyl, Tokenize, Chunks, ChunkFramework


sys.path.append("../")


def test_syl_tokenize():
    instr = " མཐའི་རྒྱ་མཚོའི་གླིང་། ཤི་བཀྲ་ཤིས་  tr བདེ་་ལེ གས། བཀྲ་ཤིས་བདེ་ལེགས་ཀཀ"
    preprocessed = TokChunks(instr)
    preprocessed.serve_syls_to_trie()
    config = Config()
    trie = Trie(BoSyl, config.profile, config.dictionary, config.adjustments)
    tok = Tokenize(trie)
    tokens = tok.tokenize(preprocessed)
    texts = [t.text for t in tokens]
    expected = [
        " མཐའི་",
        "རྒྱ་མཚོའི་",
        "གླིང་",
        "། ",
        "ཤི་",
        "བཀྲ་ཤིས་  ",
        "tr ",
        "བདེ་་ལེ གས",
        "། ",
        "བཀྲ་ཤིས་",
        "བདེ་ལེགས་",
        "ཀཀ",
    ]
    # current: [' མཐའི་', 'རྒྱ་མཚོའི་', '། ', 'གླིང་', 'བཀྲ་', 'ཤི་', 'tr ', 'ཤིས་  ', 'བདེ་་ལེ གས', '། ', 'བདེ་',
    #          'བཀྲ་ཤིས་', 'ཀཀ', 'ལེགས་']
    assert texts == expected


def test_num_lemmas_missing(wt):
    in_str = "སྟོང་ཕྲག་བརྒྱ་པ་སུམ་བརྒྱ་པ་བཅུ་པ་ལྔ་པ་"
    tokens = wt.tokenize(in_str)
    assert [t.lemma for t in tokens] == [
        "སྟོང་ཕྲག་",
        "བརྒྱ་པ་",
        "སུམ་བརྒྱ་པ་",
        "བཅུ་པ་",
        "ལྔ་པ་",
    ]


def test_no_shad_syllable():
    in_str = "ཀ འདི་ ཤི དེ་ག རེད་དོ།"
    bo_string = Chunks(in_str)
    chunks = bo_string.make_chunks()
    chunks = bo_string.get_readable(chunks)
    assert chunks == [
        ("TEXT", "ཀ "),
        ("TEXT", "འདི་ "),
        ("TEXT", "ཤི "),
        ("TEXT", "དེ་"),
        ("TEXT", "ག "),
        ("TEXT", "རེད་"),
        ("TEXT", "དོ"),
        ("PUNCT", "།"),
    ]


def test_segmentation_bug(wt):
    tokens = wt.tokenize("ལ་པོ་ལ་པོ་ལ་པོ་")
    assert len(tokens) == 3

    tokens = wt.tokenize("ལ་མོ་ལ་མོ་ལ་མོ་")
    assert len(tokens) == 3

    tokens = wt.tokenize("གྲོགས་པོ་གྲོགས་པོ་གྲོགས་པོ་")
    assert len(tokens) == 3

    tokens = wt.tokenize("བདག་པོ་བདག་པོ་བདག་པོ་དང་")
    assert len(tokens) == 4

    tokens = wt.tokenize("བདག་པོ་བདག་པོ་བདག་པོ་")
    assert len(tokens) == 3

    tokens = wt.tokenize(
        "བདག་པོ་བདག་པོ་བདག་པོ་བདག་པོ་བདག་པོ་བདག་པོ་བདག་པོ་བདག་པོ་བདག་པོ་"
    )
    assert len(tokens) == 9


def test_keyerror_part_lemma(wt):
    tokens = wt.tokenize("ཕའིའོ།")
    assert len(tokens) == 3


def test_split_token(empty_wt):
    wt = empty_wt
    wt.tok.trie.rebuild_trie()
    wt.tok.trie.inflect_n_modify_trie("འ་")
    assert not wt.tok.trie.has_word("ར་")["exists"]


def test_missing_entries_n_bad_unaffixed(wt):
    input_str = "ཤུ་ཀ་ར་"
    tokens = wt.tokenize(input_str, split_affixes=False)
    assert [t.text for t in tokens] == ["ཤུ་", "ཀ་ར་"]
    assert tokens[0].senses
    assert tokens[1].text_unaffixed == "ཀ་ར་"


def test_multiple_spaces():
    bo_string = Chunks("ཁྱོ ད་ད  ང་")
    chunks = bo_string.make_chunks()
    chunks = bo_string.get_readable(chunks)
    assert chunks[0] == ("TEXT", "ཁྱོ ད་")
    assert chunks[1] == ("TEXT", "ད  ང་")
    assert len(chunks) == 2


def test_bug1(wt):
    string = "བ་ཀུ་"
    tokens = wt.tokenize(string, debug=True)
    assert tokens


def test_bug2(wt):
    string = "བྲ་གྲྀ་"
    tokens = wt.tokenize(string, debug=True)
    assert tokens


def test_many_tseks_in_syllable():
    input_str = " ཤི་བཀྲ་ཤིས་  བདེ་་ལ             ེ       གས་ བཀྲ་ཤིས་བདེ་ལེགས"
    cb = ChunkFramework(input_str)
    chunks = cb.syllabify()
    readable = cb.get_readable(chunks)
    assert readable == [
        ("TEXT", " ཤི་"),
        ("TEXT", "བཀྲ་"),
        ("TEXT", "ཤིས་"),
        ("TEXT", "  བདེ་་"),
        ("TEXT", "ལ             ེ       གས་"),
        ("TEXT", " བཀྲ་"),
        ("TEXT", "ཤིས་"),
        ("TEXT", "བདེ་"),
        ("TEXT", "ལེགས"),
    ]

    chunks = cb.chunk_punct()
    chunks = cb.merge_skippable_punct(chunks)
    readable = cb.get_readable(chunks)
    assert readable == [
        ("NON_PUNCT", " ཤི་བཀྲ་ཤིས་  བདེ་་ལ             ེ       གས་ བཀྲ་ཤིས་བདེ་ལེགས")
    ]

    ck = Chunks(input_str)
    chunks = ck.make_chunks()
    readable = ck.get_readable(chunks)
    assert readable == [
        ("TEXT", " ཤི་"),
        ("TEXT", "བཀྲ་"),
        ("TEXT", "ཤིས་  "),
        ("TEXT", "བདེ་་"),
        ("TEXT", "ལ             ེ       གས་ "),
        ("TEXT", "བཀྲ་"),
        ("TEXT", "ཤིས་"),
        ("TEXT", "བདེ་"),
        ("TEXT", "ལེགས"),
    ]


def test_shad_in_syllable():
    input_str = " tr བདེ་་ལེ གས། བཀྲ་"
    ck = Chunks(input_str)
    chunks = ck.make_chunks()
    readable = ck.get_readable(chunks)
    assert readable == [
        ("LATIN", " tr "),
        ("TEXT", "བདེ་་"),
        ("TEXT", "ལེ གས"),
        ("PUNCT", "། "),
        ("TEXT", "བཀྲ་"),
    ]


if __name__ == "__main__":
    test_split_token()



# coding: utf-8
from ..third_party.cqlparser import Query
from .tokenmerge import TokenMerge


class MergingMatcher:
    def __init__(self, query, replace_idx, token_list, token_changes=None):
        self.matcher = Query(query)
        self.span = len(self.matcher.tokenexprs) - 1
        self.token_list = token_list

        self.replace_idx = replace_idx - 1
        self.token_changes = token_changes

    def merge_on_matches(self):
        merged_list = []
        i = 0
        while i < len(self.token_list):
            if self.__matches(i):
                # find the index of the token to split
                idx = i + self.replace_idx

                # add new tokens that precede the one to split
                for r in range(i, idx):
                    merged_list.append(self.token_list[r])
                    i += 1

                # split the token and add them to the new list
                merged_list.append(
                    self.__merge(self.token_list[idx], self.token_list[idx + 1])
                )
                i += 1
            else:
                merged_list.append(self.token_list[i])

            i += 1

        return merged_list

    def __matches(self, i):
        return i + self.span <= len(self.token_list) and self.matcher(
            self.token_list[i : i + self.span + 1]
        )

    def __merge(self, token1, token2):
        ts = TokenMerge(token1, token2, self.token_changes)
        return ts.merge()



# coding: utf-8
from pathlib import Path


class PipelineBase:
    def __init__(self, profile, pipes=None):
        self.pipes = pipes

        self.prep = None
        self.tok = None
        self.mod = None
        self.form = None

        self.left = 5
        self.right = 5
        self.tok_params = None
        self.filename = (
            None  # for an advanced mode, to show what conc comes from which file
        )

        self.args_list = {
            "prep",
            "tok",
            "mod",
            "form",  # components
            "tok_params",  # pybo
            "left",
            "right",  # concs
            "filename",
        }  # others

        self.parse_profile(profile)

    def pipe_str(self, text: str) -> str:
        # a. preprocessing
        if self.prep:
            text = self.pipes["prep"][self.prep](text)

        # b. tokenizing
        if (
            isinstance(self.tok, str)
            and (
                "word" in self.tok or "sentence" in self.tok or "paragraph" in self.tok
            )
            and self.tok_params
        ):
            elts = self.pipes["tok"][self.tok](text, config=self.tok_params["config"])
        else:
            elts = self.pipes["tok"][self.tok](text)

        # c. modifying
        mod = self.pipes["mod"][self.mod]
        if isinstance(self.mod, str) and self.mod.endswith("concs"):
            elts = mod(elts, left=self.left, right=self.right)
        else:
            elts = mod(elts)

        # d. formatting
        elts = self.pipes["form"][self.form](elts)

        return elts

    def pipe_file(self, filename: str, out_file: str):
        in_file = Path(filename)
        out_file = Path(out_file)
        assert in_file.is_file()

        with in_file.open(encoding="utf-8-sig") as f:
            dump = f.read()

        output = self.pipe_str(dump)

        with out_file.open("w", encoding="utf-8-sig") as g:
            g.write(output)

    def parse_profile(self, pipeline):
        self.is_valid_params(pipeline)
        for arg, v in pipeline.items():
            if arg == "prep":
                self.prep = v
            elif arg == "tok":
                self.tok = v
            elif arg == "mod":
                self.mod = v
            elif arg == "form":
                self.form = v
            elif arg == "tok_params":
                self.tok_params = v
            elif arg == "left":
                self.left = v
            elif arg == "right":
                self.right = v
            elif arg == "filename":
                self.filename = v
        self.is_valid_pipeline()

    def is_valid_params(self, pipeline):
        for arg, val in pipeline.items():
            # ensure all arguments are valid attributes
            if arg not in self.args_list:
                raise SyntaxError(
                    f'{arg} is not a valid argument\nvalid options are {" ".join(self.map)}'
                )

            # ensure arguments have valid values
            if arg in self.pipes and val not in self.pipes[arg]:
                raise SyntaxError(
                    f'{val} is not a valid value for {arg}\nvalid options are {" ".join(self.pipes[arg])}'
                )

    def is_valid_pipeline(self):
        # missing pipes
        if not self.tok or not self.mod or not self.form:
            raise BrokenPipeError(
                "A valid pipeline must have a tokenizer, a processor and a formatter."
            )



# coding: utf-8
from .tokenmerge import TokenMerge
from ..vars import TSEK, DAGDRA


class MergeDagdra:
    """
    A class to merge pa/po/ba/bo tokens in a token list produced by BoTokenizer

    """

    def __init__(self):
        pass

    def merge(self, tokens):
        """
        Merges the tokens containing either pa/po/ba/bo

        :param tokens: list of Token objects
        """
        if len(tokens) <= 1:
            pass
        elif len(tokens) == 2:
            token0, token1 = tokens
            if token1.text_cleaned in DAGDRA:
                # split token containing the affixed particle
                merged = self.merge_with_previous_token(token0, token1)
                del tokens[1]
                tokens[0] = merged
        else:
            t = 0
            while t <= len(tokens) - 1:
                if t + 1 > len(tokens) - 1:
                    break
                token0, token1 = tokens[t], tokens[t + 1]
                clean_word = (
                    token1.text_cleaned + TSEK
                    if not token1.text_cleaned.endswith(TSEK)
                    else token1.text_cleaned
                )
                if (
                    token0.chunk_type == "TEXT"
                    and token1.chunk_type == "TEXT"
                    and clean_word in DAGDRA
                ):
                    # split token containing the affixed particle
                    merged = self.merge_with_previous_token(token0, token1)

                    # replace the original token with the two new ones
                    tokens[t : t + 2] = [merged]
                t += 1

    def merge_with_previous_token(self, token0, token1):
        merged = TokenMerge(token0, token1).merge()
        merged.has_merged_dagdra = True
        merged.lemma = merged.text_cleaned
        return merged



# coding: utf-8
from .sylcomponents import SylComponents


class BoSyl(SylComponents):
    def __init__(self):
        SylComponents.__init__(self)
        self.affixes = {
            "ར": {"len": 1, "type": "la"},
            "ས": {"len": 1, "type": "gis"},
            "འི": {"len": 2, "type": "gi"},
            "འམ": {"len": 2, "type": "am"},
            "འང": {"len": 2, "type": "ang"},
            "འོ": {"len": 2, "type": "o"},
            "འིའོ": {"len": 4, "type": "gi+o"},
            "འིའམ": {"len": 4, "type": "gi+am"},
            "འིའང": {"len": 4, "type": "gi+ang"},
            "འོའམ": {"len": 4, "type": "o+am"},
            "འོའང": {"len": 4, "type": "o+ang"},
        }

    def is_affixable(self, syl):
        """expects a clean syllable without ending tsek"""
        affixable = False
        if self.is_thame(syl):
            affixable = True
            for ending in ["ར", "ས", "འི", "འོ", "མ", "ང"]:
                if len(syl) > len(ending) and syl.endswith(ending):
                    affixable = False
        return affixable

    def get_all_affixed(self, syl):
        """
        :param syl: syl to be affixed
        :return: if affixable: [(<syl+affixed>, {'len': int, 'type': str, 'aa': bool}), (..., ...)]
                 otherwise   : <syl>
        """
        if self.is_affixable(syl):
            aa = False
            if syl.endswith("འ") and len(syl) > 1:
                syl = syl[:-1]
                aa = True

            affixed = []
            for a in self.affixes.keys():
                metadata = {}
                metadata.update(self.affixes[a])
                metadata.update({"aa": aa})
                affixed.append((syl + a, metadata))
            return affixed

        else:
            return None



#
# Original copyright notice:
#
# ---------------------------------------------------------------
# PyNLPl - Corpus Query Language (CQL)
#   by Maarten van Gompel
#   Centre for Language Studies
#   Radboud University Nijmegen
#   http://proycon.github.com/folia
#   http://www.github.com/proycon/pynlpl
#   proycon AT anaproy DOT nl
#
# Parser and interpreter for a basic subset of the Corpus Query Language
#
#   Licensed under GPLv3
#
# ----------------------------------------------------------------
#
# This file is modified and reditstributed here under APL2 with
# with written permission from the original author

from __future__ import print_function, unicode_literals, division, absolute_import

from .fsa import State, NFA
import re
import sys

OPERATORS = ("=", "!=")
MAXINTERVAL = 99
DEFAULTFIELD = "text_cleaned"  # pynpl default: "word"


class SyntaxError(Exception):
    pass


class ValueExpression(object):
    def __init__(self, values):
        self.values = values  # disjunction

    @staticmethod
    def parse(s, i):
        values = ""
        assert s[i] == '"'
        i += 1
        while not (s[i] == '"' and s[i - 1] != "\\"):
            values += s[i]
            i += 1
        values = values.split("|")
        return ValueExpression(values), i + 1

    def __len__(self):
        return len(self.values)

    def __iter__(self):
        for x in self.values:
            yield x

    def __getitem__(self, index):
        return self.values[index]


class AttributeExpression(object):
    def __init__(self, attribute, operator, valueexpression):
        self.attribute = attribute
        self.operator = operator
        self.valueexpr = valueexpression

    @staticmethod
    def parse(s, i):
        while s[i] == " ":
            i += 1
        if s[i] == '"':
            # no attribute and no operator, use defaults:
            attribute = DEFAULTFIELD  # pynpl default: "word"
            operator = "="
        else:
            attribute = ""
            while s[i] not in (" ", "!", ">", "<", "="):
                attribute += s[i]
                i += 1
            if not attribute:
                raise SyntaxError("Expected attribute name, none found")
            operator = ""
            while s[i] in (" ", "!", ">", "<", "="):
                if s[i] != " ":
                    operator += s[i]
                i += 1
            if operator not in OPERATORS:
                raise SyntaxError("Expected operator, got '" + operator + "'")
        if s[i] != '"':
            raise SyntaxError(
                "Expected start of value expression (doublequote) in position "
                + str(i)
                + ", got "
                + s[i]
            )
        valueexpr, i = ValueExpression.parse(s, i)
        return AttributeExpression(attribute, operator, valueexpr), i


class TokenExpression(object):
    def __init__(self, attribexprs=[], interval=None):
        self.attribexprs = attribexprs
        self.interval = interval

    @staticmethod
    def parse(s, i):
        attribexprs = []
        while s[i] == " ":
            i += 1
        if s[i] == '"':
            attribexpr, i = AttributeExpression.parse(s, i)
            attribexprs.append(attribexpr)
        elif s[i] == "[":
            i += 1
            while True:
                while s[i] == " ":
                    i += 1
                if s[i] == "&":
                    attribexpr, i = AttributeExpression.parse(s, i + 1)
                    attribexprs.append(attribexpr)
                elif s[i] == "]":
                    i += 1
                    break
                elif not attribexprs:
                    attribexpr, i = AttributeExpression.parse(s, i)
                    attribexprs.append(attribexpr)
                else:
                    raise SyntaxError(
                        "Unexpected char whilst parsing token expression,  position "
                        + str(i)
                        + ": "
                        + s[i]
                    )
        else:
            raise SyntaxError(
                'Expected token expression starting with either " or [, got: ' + s[i]
            )

        if i == len(s):
            interval = None  # end of query!
        elif s[i] == "{":
            # interval expression, find end:
            interval = None
            for j in range(i + 1, len(s)):
                if s[j] == "}":
                    interval = s[i + 1 : j]

            if interval is None:
                raise SyntaxError("Interval expression started but no end-brace found")

            i += len(interval) + 2

            try:
                if "," in interval:
                    interval = tuple(int(x) for x in interval.split(","))
                    if len(interval) != 2:
                        raise SyntaxError("Invalid interval: " + interval)
                elif "-" in interval:  # alternative
                    interval = tuple(int(x) for x in interval.split("-"))
                    if len(interval) != 2:
                        raise SyntaxError("Invalid interval: " + interval)
                else:
                    interval = (int(interval), int(interval))
            except ValueError:
                raise SyntaxError("Invalid interval: " + interval)
        elif s[i] == "?":
            interval = (0, 1)
            i += 1
        elif s[i] == "+":
            interval = (1, MAXINTERVAL)
            i += 1
        elif s[i] == "*":
            interval = (0, MAXINTERVAL)
            i += 1
        else:
            interval = None

        return TokenExpression(attribexprs, interval), i

    def __len__(self):
        return len(self.attribexprs)

    def __iter__(self):
        for x in self.attribexprs:
            yield x

    def __getitem__(self, index):
        return self.attribexprs[index]

    def nfa(self, nextstate):
        """Returns an initial state for an NFA"""
        if self.interval:
            mininterval, maxinterval = (
                self.interval
            )  # pylint: disable=unpacking-non-sequence
            nextstate2 = nextstate
            for i in range(maxinterval):
                state = State(transitions=[(self, self.match, nextstate2)])
                if i + 1 > mininterval:
                    if nextstate is not nextstate2:
                        state.transitions.append((self, self.match, nextstate))
                    if maxinterval == MAXINTERVAL:
                        state.epsilon.append(state)
                        break
                nextstate2 = state
            return state
        else:
            state = State(transitions=[(self, self.match, nextstate)])
            return state

    def match(self, value):
        match = False
        for _, attribexpr in enumerate(self):
            annottype = attribexpr.attribute
            # if annottype == 'text': annottype = 'word'  # commented because pybo's text attribute is "text"
            if attribexpr.operator == "!=":
                negate = True
            elif attribexpr.operator == "=":
                negate = False
            else:
                raise Exception("Unexpected operator " + attribexpr.operator)

            if len(attribexpr.valueexpr) > 1:
                expr = re.compile("^(" + "|".join(attribexpr.valueexpr) + ")$")
            else:
                expr = re.compile("^" + attribexpr.valueexpr[0] + "$")
            match = expr.match(value[annottype]) is not None
            if negate:
                match = not match
            if not match:
                return False
        return True


class Query(object):
    def __init__(self, s):
        self.tokenexprs = []
        i = 0
        l = len(s)
        while i < l:
            if s[i] == " ":
                i += 1
            else:
                tokenexpr, i = TokenExpression.parse(s, i)
                self.tokenexprs.append(tokenexpr)

    def __len__(self):
        return len(self.tokenexprs)

    def __iter__(self):
        for x in self.tokenexprs:
            yield x

    def __getitem__(self, index):
        return self.tokenexprs[index]

    def nfa(self):
        """convert the expression into an NFA"""
        finalstate = State(final=True)
        nextstate = finalstate
        for tokenexpr in reversed(self):
            state = tokenexpr.nfa(nextstate)
            nextstate = state
        return NFA(state)

    def __call__(self, tokens, debug=False):
        """Execute the CQL expression, pass a list of tokens/annotations using keyword arguments: word, pos, lemma, etc"""

        if not tokens:
            raise Exception(
                "Pass a list of tokens/annotation using keyword arguments! (word,pos,lemma, or others)"
            )

        # convert the expression into an NFA
        nfa = self.nfa()
        if debug:
            print(repr(nfa), file=sys.stderr)

        return list(nfa.find(tokens, debug))



# coding: utf8

# variables used in the tests
ending_particles = [
    "གོ་",
    "ངོ་",
    "དོ་",
    "ནོ་",
    "བོ་",
    "མོ་",
    "འོ་",
    "རོ་",
    "ལོ་",
    "སོ་",
    "ཏོ་",
]
ending_words = ["ཅིག་", "ཤོག་"]
ending_verbs = [
    "ཡིན་",
    "ཡོད་",
    "མིན་",
    "མེད་",
    "འགྱུར་",
    "ལྡན་",
    "བགྱི་",
    "བྱ་",
    "བཞུགས་",
]
te_particles = [
    "སྟེ་",
    "ཏེ་",
    "དེ་",
]  # separated because these seem to cut long sentences
clause_boundaries = te_particles + ["ནས་", "ན་"]
dagdra = ["པ་", "བ་", "པོ་", "བོ་"]


# Turn sentences as indices into sentences as follows:
# (<sent_length>, [token1, token2, ..., tokenn]) where tokens are Token objects
############################################################
def sentence_tokenizer(tokens):
    sent_indices = get_sentence_indices(tokens)
    # get tokens for each sentence
    sentences = []
    for sentence in sent_indices:
        start, end, l = sentence["start"], sentence["end"], sentence["len"]
        sentences.append((l, tokens[start : end + 1]))

    return sentences


# Get sentences in the form of indices
############################################################
def get_sentence_indices(tokens):
    """
    from 1 to 3, we want to use the shad since it reflects the writer's own idea of where parts of the
    sentence end and start.
    1. ending particle + shad to not take all the quotations that potentially contain ending particles

    2. clause boundary particles + shad: ན་ is often found within a sentence, but cuts it in a way that doesn't prevent
    the correct understanding, so for translation, we might want to remove it as translation units will want
    to keep those sentences together, but for segmentation jobs, it works as a fairly safe segmenter.
    we might want to use it.

    Output: list of sentences, each in the following format: (sentence-length, [word1, word2, ...])
    """
    # 1. find unambiguous sentence end markers: ending particles followed by punctuation
    previous_end = 0
    sentence_idx = extract_chunks(is_endpart_n_punct, tokens, 0, previous_end)

    # 2. find clause boundaries followed by punctuation
    sentence_idx = piped_sentencify(sentence_idx, tokens, is_clause_boundary_n_punct)

    # 3. find verbs followed by punctuation
    sentence_idx = piped_sentencify(sentence_idx, tokens, is_verb_n_punct)

    # 4. find verbs followed by clause boundaries
    sentence_idx = piped_sentencify(
        sentence_idx, tokens, is_verb_n_clause_boundary, threshold=30
    )  # max size to check

    # joining the sentences without verbs to either the one preceding them or following them
    sentence_idx = join_no_verb_sentences(sentence_idx, tokens)

    return sentence_idx


# Tests used in the sentence tokenizer
############################################################
def is_word(token):
    return token and token.chunk_type == "TEXT"


def has_last_syl(token, l):
    if token.syls:
        last_syl = "".join(token.syls[-1]) + "་"
        return last_syl in l
    else:
        return False


def is_ending_part(token):
    return token and token.pos == "PART" and has_last_syl(token, ending_particles)


def is_endpart_n_punct(token1, token2):
    return is_ending_part(token1) and token2.chunk_type == "PUNCT"


def is_clause_boundary_n_punct(token1, token2):
    return (
        has_last_syl(token1, clause_boundaries) or has_last_syl(token1, ending_words)
    ) and token2.chunk_type == "PUNCT"


def is_verb_n_punct(token1, token2):
    return (
        (token1.pos == "VERB" and not has_last_syl(token1, dagdra))
        or has_last_syl(token1, ending_verbs)
    ) and token2.chunk_type == "PUNCT"


def is_verb_n_clause_boundary(token1, token2):
    return (
        (token1.pos == "VERB" and not has_last_syl(token1, dagdra))
        or has_last_syl(token1, ending_verbs)
    ) and has_last_syl(token2, clause_boundaries)


# sentence tokenizer logic
############################################################
def join_no_verb_sentences(sentences, tokens, threshold=4):
    i = 0
    while i < len(sentences):
        start, end, length = (
            sentences[i]["start"],
            sentences[i]["end"],
            sentences[i]["len"],
        )
        if length > threshold:
            i += 1
            continue

        no_verb = True
        for token in tokens[start : end + 1]:
            if token.pos == "VERB" and not has_last_syl(token, dagdra):
                no_verb = False

        if no_verb:
            last_word = (
                tokens[end] if tokens[end].chunk_type == "TEXT" else tokens[end - 1]
            )
            if i + 1 < len(sentences) and has_last_syl(last_word, clause_boundaries):
                # join to the left
                sentences[i + 1]["start"] = sentences[i]["start"]
                sentences[i + 1]["len"] += sentences[i]["len"]
                del sentences[i]
                i -= 1
            elif i - 1 >= 0:
                # join to the right
                last_word_of_previous = (
                    tokens[sentences[i - 1]["end"]]
                    if tokens[sentences[i - 1]["end"]].chunk_type == "TEXT"
                    else tokens[sentences[i - 1]["end"] - 1]
                )
                if not has_last_syl(last_word_of_previous, ending_particles):
                    sentences[i - 1]["end"] = sentences[i]["end"]
                    sentences[i - 1]["len"] += sentences[i]["len"]
                    del sentences[i]
                    i -= 1

        i += 1

    return sentences


def extract_chunks(test, subtokens, start, previous_end):
    chunks = []
    n = 0
    for n, token in enumerate(subtokens):
        if test(subtokens[n - 1], token):
            chunks.append(
                {
                    "start": previous_end,
                    "end": start + n,
                    "len": start + n + 1 - previous_end,
                }
            )
            previous_end = start + n + 1
    if chunks and previous_end < start + n:
        chunks.append(
            {
                "start": previous_end,
                "end": start + n,
                "len": start + n + 1 - previous_end,
            }
        )

    # add all the subtokens if no chunk was produced
    if not chunks:
        chunks.append(
            {
                "start": previous_end,
                "end": start + n,
                "len": start + n + 1 - previous_end,
            }
        )

    return chunks


def piped_sentencify(sentences, tokens, test, threshold=None):
    a = 0
    while a < len(sentences):
        length, start, end = (
            sentences[a]["len"],
            sentences[a]["start"],
            sentences[a]["end"],
        )
        previous_end = start
        new_sentences = []
        if threshold:
            if end - start > threshold:
                new_sentences = extract_chunks(
                    test, tokens[start : end + 1], start, previous_end
                )
        else:
            new_sentences = extract_chunks(
                test, tokens[start : end + 1], start, previous_end
            )

        if new_sentences:
            sentences[a : a + 1] = new_sentences
            a += len(new_sentences)
        else:
            a += 1
    return sentences



# coding: utf-8
import copy

from ..third_party.cqlparser import replace_token_attributes


class TokenSplit:
    """
    Takes a token object and divide it into two using an index of the content.

    The affected attributes are:
        - token.content     : the string is split at the index
        - token.char_groups : the dict items are redistributed
        - token.start       : second token only. now equals "start + index"
        - token.len      : length of new content
        - token.syls        : syls are redistributed and split if necessary

    """

    def __init__(self, token, split_idx, token_changes=None):
        self.token = token
        self.first = None
        self.second = None
        self.token_changes = token_changes
        self.idx = split_idx

    def split(self, mode="char"):
        """
        :param mode: can either be "syl" or "char" to split on a syllable index or a character index.
        """
        if mode != "char" and mode != "syl":
            raise SyntaxError("splitting mode should either be 'syl' or 'char'. ")

        # in syllable-mode, if there is only one syllable, return the word without splitting it.
        if mode == "syl" and len(self.token.syls) == 1:
            return [self.token]

        self.split_on_idx(mode=mode)
        self.replace_attrs()

        return [self.first, self.second]

    def replace_attrs(self):
        if self.token_changes:
            tokens = [self.first, self.second]
            replace_token_attributes(tokens, self.token_changes)
            self.first, self.second = tokens

    def split_on_idx(self, mode):
        self.first = copy.deepcopy(self.token)
        self.second = copy.deepcopy(self.token)

        if mode == "syl":
            self.idx = self.token.syls_start_end[self.idx - 1]["end"]

        self.__split_contents()
        self.__split_indices()
        self.__split_syls_idx()
        self.__split_syls_start_end(mode)
        self.__split_char_types()
        self.__split_affixation()

    def __split_contents(self):
        text = self.first.text
        self.first.text = text[0 : self.idx]
        self.second.text = text[self.idx :]

    def __split_char_types(self):
        char_types = self.first.char_types
        self.first.char_types = char_types[: self.idx]
        self.second.char_types = char_types[self.idx :]

    def __split_indices(self):
        self.first.len = len(self.first.text)
        self.second.len = len(self.second.text)
        self.second.start = self.second.start + self.idx

    def __split_syls_start_end(self, mode):
        if not self.token.syls_start_end:
            return

        to_split_idx = 0
        for num, s in enumerate(self.token.syls_start_end):
            if s["start"] <= self.idx <= s["end"]:
                to_split_idx = num
                break  # ensure to exit on first match

        start = self.token.syls_start_end[:to_split_idx]
        end = self.token.syls_start_end[to_split_idx + 1 :]
        to_split = self.token.syls_start_end[to_split_idx]

        if mode == "char":
            start.append({"start": to_split["start"], "end": self.idx})
            end.append({"start": self.idx, "end": to_split["end"]})

        if mode == "syl":
            start.append(to_split)

        self.first.syls_start_end = start
        self.second.syls_start_end = end

    def __split_syls_idx(self):
        syls = self.first.syls_idx
        # empty syls
        self.first.syls_idx = []
        self.second.syls_idx = []

        if syls:
            for syl in syls:
                if syl[-1] < self.idx:
                    self.first.syls_idx.append(syl)

                else:
                    # separate the syl in two
                    part1, part2 = [], []
                    for i in syl:
                        if i < self.idx:
                            part1.append(i)
                        else:
                            part2.append(i - self.idx)

                    # add them if non-empty
                    if part1:
                        self.first.syls_idx.append(part1)
                    if part2:
                        self.second.syls_idx.append(part2)

    def __split_affixation(self):
        if self.token.affixation:
            del self.first.affixation["len"]
            del self.first.affixation["type"]
            del self.second.affixation["aa"]



# coding: utf-8
from ..third_party.cqlparser import Query, replace_token_attributes


class ReplacingMatcher:
    def __init__(self, query, replace_idx, token_list, token_changes=None):
        self.matcher = Query(query)
        self.span = len(self.matcher.tokenexprs) - 1
        self.replace_idx = replace_idx - 1
        self.token_list = token_list
        self.token_changes = token_changes

    def replace_on_matches(self):
        i = 0
        while i < len(self.token_list):
            if self.__matches(i):
                # find the index of the token to split
                idx = i + self.replace_idx
                replace_token_attributes(self.token_list[idx], self.token_changes)
            i += 1

    def __matches(self, i):
        return i + self.span <= len(self.token_list) and self.matcher(
            self.token_list[i : i + self.span + 1]
        )



# coding: utf-8
import copy

from ..third_party.cqlparser import replace_token_attributes


class TokenMerge:
    """

    """

    def __init__(self, token1, token2, token_changes=None):
        self.token1 = token1
        self.token2 = token2
        self.merged = copy.deepcopy(token1)
        if not self.merged.syls_idx:
            self.merged.syls_idx = []
        if not self.merged.syls:
            self.merged.syls = []
        self.token_changes = token_changes

    def merge(self):
        self.merge_attrs()
        self.replace_attrs()
        return self.merged

    def replace_attrs(self):
        """
        Replaces the content of attributes that were found in the cql query.
        If no query is provided, the values of the first token are kept.
        """
        if self.token_changes:
            replace_token_attributes(self.merged, self.token_changes)

    def merge_attrs(self):
        self.__merge_texts()
        self.__merge_indices()
        self.__merge_syls_idx()
        self.__merge_syls_start_end()
        self.__del_lemma()

    def __merge_texts(self):
        self.merged.text += self.token2.text

    def __merge_indices(self):
        self.merged.len += self.token2.len

    def __merge_syls_start_end(self):
        # token1 is a host syllable and token2 its affixed syllable
        if (
            not self.merged.syls_start_end
            or not self.token1.syls_start_end
            or not self.token2.syls_start_end
        ):
            return

        if (
            self.token1.affix_host
            and not self.token1.affix
            and not self.token2.affix_host
            and self.token2.affix
        ):
            self.merged.syls_start_end[-1]["end"] = self.token2.syls_start_end[0]["end"]
            self.merged.syls_start_end.extend(self.token2.syls_start_end[1:])
        else:
            self.merged.syls_start_end.extend(self.token2.syls_start_end)

    def __merge_syls_idx(self):
        """
        Updates indices and add the syls to the merged object
        Re-joins the host-syllable and affixed particle syllables into a single one;
        then, affix is True and affixed also, so cleaned_content gets its tsek.
        """
        first_syl = True
        if self.token2.syls_idx:
            for syl in self.token2.syls_idx:
                if syl:
                    new_syl = [i + self.token1.len for i in syl]

                    # token1 is a host syllable and token2 its affixed syllable
                    if (
                        first_syl
                        and (self.token1.affix_host and not self.token1.affix)
                        and (not self.token2.affix_host and self.token2.affix)
                    ):
                        self.merged.syls_idx[-1] += new_syl
                        self.merged.affix = True
                        first_syl = False
                    else:
                        self.merged.syls_idx.append(new_syl)

    def __del_lemma(self):
        """
        Simply deletes any lemma in merged since the lemma of the merged token can't be guessed.
        """
        if self.token1["lemma"]:
            self.merged.lemma = None



#
# Original copyright notice:
#
# ---------------------------------------------------------------
# PyNLPl - Finite State Automata
#   by Maarten van Gompel
#   Centre for Language Studies
#   Radboud University Nijmegen
#   http://proycon.github.com/folia
#   http://www.github.com/proycon/pynlpl
#   proycon AT anaproy DOT nl
#
# Partially based/inspired on code by Xiayun Sun (https://github.com/xysun/regex)
#
#   Licensed under GPLv3
#
# ----------------------------------------------------------------
#
# This file is modified and reditstributed here under APL2 with
# with written permission from the original author

from __future__ import print_function, unicode_literals, division, absolute_import
import sys


class State(object):
    def __init__(self, **kwargs):
        if "epsilon" in kwargs:
            self.epsilon = kwargs["epsilon"]  # epsilon-closure (lis of states)
        else:
            self.epsilon = []  # epsilon-closure
        if "transitions" in kwargs:
            self.transitions = kwargs["transitions"]
        else:
            self.transitions = []  # (matchitem, matchfunction(value), state)
        if "final" in kwargs:
            self.final = bool(kwargs["final"])  # ending state
        else:
            self.final = False
        self.transitioned = (
            None
        )  # will be a tuple (state, matchitem) indicating how this state was reached


class NFA(object):
    """Non-deterministic finite state automaton. Can be used to model DFAs as well if your state transitions are not ambiguous and epsilon is empty."""

    def __init__(self, initialstate):
        self.initialstate = initialstate

    def run(self, sequence, mustmatchall=False, debug=False):
        def add(state, states):
            """add state and recursively add epsilon transitions"""
            assert isinstance(state, State)
            if state in states:
                return
            states.add(state)
            for eps in state.epsilon:  # recurse into epsilon transitions
                add(eps, states)

        current_states = set()
        add(self.initialstate, current_states)
        if debug:
            print(
                "Starting run, current states: ", repr(current_states), file=sys.stderr
            )

        for offset, value in enumerate(sequence):
            if not current_states:
                break
            if debug:
                print("Value: ", repr(value), file=sys.stderr)
            next_states = set()
            for state in current_states:
                for matchitem, matchfunction, trans_state in state.transitions:
                    if matchfunction(value):
                        trans_state.transitioned = (state, matchitem)
                        add(trans_state, next_states)

            current_states = next_states
            if debug:
                print("Current states: ", repr(current_states), file=sys.stderr)
            if not mustmatchall:
                for s in current_states:
                    if s.final:
                        if debug:
                            print("Final state reached", file=sys.stderr)
                        yield offset + 1

        if mustmatchall:
            for s in current_states:
                if s.final:
                    if debug:
                        print("Final state reached", file=sys.stderr)
                    yield offset + 1

    def match(self, sequence):
        try:
            return next(self.run(sequence, True)) == len(sequence)
        except StopIteration:
            return False

    def find(self, sequence, debug=False):
        l = len(sequence)
        for i in range(0, l):
            for length in self.run(sequence[i:], False, debug):
                yield sequence[i : i + length]



# coding: utf-8
import re


def basic_cleanup(text: str) -> str:
    text = text.strip()
    text = re.sub(r"\n+", " ", text)
    text = re.sub(r"\s+", " ", text)
    return text


def basic_keeps_lines(text: str) -> str:
    text = text.strip()
    # text = re.sub(r'\s+', ' ', text)
    return text



# coding: utf-8
from functools import lru_cache
from typing import List, NewType, Tuple

from ..tokenizers.chunktokenizer import ChunkTokenizer
from ..tokenizers.paragraphtokenizer import paragraph_tokenizer
from ..tokenizers.sentencetokenizer import sentence_tokenizer
from ..tokenizers.token import Token
from ..tokenizers.wordtokenizer import WordTokenizer

BoToken = NewType("BoToken", Token)


def space_tok(text: str) -> List[str]:
    """Tokenizes string on spaces

    """
    return text.split(" ")


def word_tok(text: str, config=None) -> List[BoToken]:
    tok = get_wordtokenizer(config=config)
    return tok.tokenize(text)


def sentence_tok(text: str, config=None) -> List[Tuple[int, List[BoToken]]]:
    tok = get_wordtokenizer(config=config)
    tokens = tok.tokenize(text)
    return sentence_tokenizer(tokens)


def paragraph_tok(text: str, config=None) -> List[Tuple[int, List[BoToken]]]:
    tok = get_wordtokenizer(config=config)
    tokens = tok.tokenize(text)
    return paragraph_tokenizer(tokens)


@lru_cache(
    maxsize=None
)  # <--- make sure that the trie is only built once then kept in memory
def get_wordtokenizer(config=None):
    return WordTokenizer(config=config)


def chunk_tok(text: str) -> List[str]:
    return ChunkTokenizer(text).tokenize()



# coding: utf-8

# inspired from https://gist.github.com/nickstanisha/733c134a0171a00f66d4
# and           https://github.com/eroux/tibetan-phonetics-py


class Node:
    def __init__(self, label=None, leaf=False, data=None):
        if data is None:
            data = {'_': {}}  # the dict in '_' is for user-data
        self.label = label
        self.leaf = leaf
        self.data = data
        self.children = dict()

    def add_child(self, key, leaf=False):
        if not isinstance(key, Node):
            self.children[key] = Node(key, leaf)
        else:
            self.children[key.leaf] = key

    def can_walk(self):
        return self.children != dict()

    def is_match(self):
        return self.leaf

    def __getitem__(self, key):
        return self.children[key]


class BasicTrie:
    def __init__(self):
        self.head = Node()

    def __getitem__(self, key):
        return self.head.children[key]

    def add(self, word, data=None):
        # adding the word
        current_node = self.head
        word_finished = True

        i = 0
        for i in range(len(word)):
            if word[i] in current_node.children:
                current_node = current_node.children[word[i]]
            else:
                word_finished = False
                break

        if not word_finished:
            while i < len(word):
                current_node.add_child(word[i])
                current_node = current_node.children[word[i]]
                i += 1

        current_node.leaf = True

        # adding data to the node
        if data:
            assert isinstance(data, dict)
            current_node.data.update(data)

    def walk(self, char, current_node=None):
        # logic of walking the trie adapted to be done outside the trie class (for Tokenize)
        if not current_node:
            current_node = self.head

        if char in current_node.children:
            next_node = current_node[char]
        else:
            next_node = None

        return next_node

    def has_word(self, word):
        if not word:
            raise ValueError('"word" must be non-null string')

        # parse the word
        current_node = self.head
        exists = True
        for syl in word:
            if syl in current_node.children:
                current_node = current_node.children[syl]
            else:
                exists = False
                break
        else:
            # reached a word like 't', not a full word in our dictionary
            if exists and not current_node.leaf:
                exists = False

        if exists:
            return {"exists": exists, "data": current_node.data}
        else:
            return {"exists": exists, "data": current_node.data}

    def add_data(self, word, data):
        """Adds data to words.

        :param word: word to add
        :param data: dict of content to add
        :return: True if any content added, False otherwise
        """
        if not word:
            raise ValueError('"word" must be non-null string')

        # parse word
        current_node = self.head
        for syl in word:
            if syl in current_node.children:
                current_node = current_node.children[syl]
            else:
                return False

        # not a complete word
        if not current_node.leaf:
            return False

        # adding data
        if isinstance(data, int):
            current_node.data["form_freq"] = data
            added = True
        else:
            if "senses" not in current_node.data:
                current_node.data["senses"] = []
            added = self.add_meaning(current_node.data["senses"], data)
        return added

    def add_meaning(self, meanings, meaning):
        if meanings:
            for m in meanings:
                if self.is_diff_meaning(meaning, m):
                    meanings.append(meaning)
                    return True
            return False
        else:
            meanings.append(meaning)
            return True

    @staticmethod
    def is_diff_meaning(m1, m2):
        is_diff = False
        for k, v in m1.items():
            if k not in m2 or k in m2 and m2[k] != v:
                is_diff = True
        return is_diff

    def deactivate(self, word, rev=False):
        """Makes word not findable (words are found only when the leaf value is True)

        :param word: word to deactivate
        :param rev: reverse the deactivation
        :return True if the word exists, False otherwise
        """
        current_node = self.head
        for syl in word:
            if syl in current_node.children:
                current_node = current_node.children[syl]
            else:
                return False
        if isinstance(current_node.data, dict):
            if not rev:
                current_node.leaf = False
            else:
                current_node.leaf = True
            return True
        else:
            return False



# coding: utf-8
import csv

from .splittingmatcher import SplittingMatcher
from .mergingmatcher import MergingMatcher
from .replacingmatcher import ReplacingMatcher
from ..utils.helpers import decomment_file


class AdjustTokens:
    """
    Syntax for the .tsv adjustment rules
    ===================================
    - each rule should be as follows: "<matchcql>\t<index>\t<operation>\t<replacecql>"
    - comments with # and empty lines are allowed
    - CQL rules: "<text>" can be used without specifying that there is "text_cleaned="
    - Index format: either "<matching_index>" or "<matching_index>-<splitting-index>"
    - Adjustment format:
            - "+" for merge
            - ":" for split (default: syllable mode)
            - "::" for split in character mode
            - "=" for replace
    - Constraint: "<matching_index>-<splitting-index>" is only allowed if adjustment is ":" or "::"
    """

    def __init__(self, main=None, custom=None):
        self.paths = []
        if custom:
            self.paths.extend(custom)
        elif main:
            self.paths.extend(main)
        self.rules = []
        self.parse_rules()

    def adjust(self, token_list):
        for rule in self.rules:
            if rule["operation"] == "split":
                sm = SplittingMatcher(
                    rule["matchcql"],
                    rule["matchidx"],
                    rule["splitidx"],
                    token_list,
                    rule["replacecql"],
                )
                token_list = sm.split_on_matches(mode=rule["splitmode"])
            elif rule["operation"] == "merge":
                mm = MergingMatcher(
                    rule["matchcql"], rule["matchidx"], token_list, rule["replacecql"]
                )
                token_list = mm.merge_on_matches()
            elif rule["operation"] == "repl":
                rm = ReplacingMatcher(
                    rule["matchcql"], rule["matchidx"], token_list, rule["replacecql"]
                )
                rm.replace_on_matches()
        return token_list

    def parse_rules(self):
        """
        Files are sorted before being applied. Thus, filenames
        :return:
        """
        for rule_file in sorted(self.paths):
            for rule in csv.reader(
                decomment_file(rule_file.open(encoding="utf-8-sig")), delimiter="\t"
            ):
                self.rules.append(self.parse_rule(rule))

    @staticmethod
    def parse_rule(rule):
        idx_sep = "-"

        # sanity checks
        if len(rule) != 4:
            raise SyntaxError("There can't be more than three columns per rule.")
        if not rule[1]:
            raise SyntaxError("There needs to be an index for every rule.")
        if idx_sep in rule[1] and rule[2] not in [":", "::"]:
            raise SyntaxError(
                "The double index in only intended for split adjustments."
            )
        if rule[2] not in ["+", "=", ":", "::"]:
            raise SyntaxError(
                'The supported operations are either of ["+", "=", ":", "::"].'
            )

        # parse
        rule_dict = {
            "matchcql": None,
            "matchidx": None,
            "operation": None,
            "splitidx": None,
            "splitmode": None,
            "replacecql": None,
        }
        rule_dict["matchcql"] = rule[0]
        if idx_sep in rule[1]:
            match_idx, split_idx = rule[1].split("-")
            rule_dict["matchidx"] = int(match_idx)
            rule_dict["splitidx"] = int(split_idx)
        else:
            rule_dict["matchidx"] = int(rule[1])
        if rule[2] == "=":
            rule_dict["operation"] = "repl"
        elif rule[2] == "+":
            rule_dict["operation"] = "merge"
        elif rule[2] == ":":
            rule_dict["operation"] = "split"
            rule_dict["splitmode"] = "syl"
        elif rule[2] == "::":
            rule_dict["operation"] = "split"
            rule_dict["splitmode"] = "char"
        rule_dict["replacecql"] = rule[3]

        return rule_dict



# coding: utf-8
from pathlib import Path
from types import FunctionType

from ..vars import Ids
from .format import *
from .modify import *
from .pipelinebase import PipelineBase
from .preprocess import *
from .tokenize import *

builtin_pipes = {
    # a. Preprocessing
    "prep": {
        "dummy": lambda x: x,
        "basic_cleanup": basic_cleanup,
        "basic_keeps_lines": basic_keeps_lines,
    },
    # b. Tokenizers
    "tok": {
        "space_tok": space_tok,
        "word_tok": word_tok,
        "chunk_tok": chunk_tok,
        "sentence_tok": sentence_tok,
        "paragraph_tok": paragraph_tok,
    },
    # c. Modifiers
    "mod": {
        "dummy": lambda x: x,
        "words_raw_text": words_raw_text,
        "words_raw_types": words_raw_types,
        "words_error_types": words_error_types,
        "words_error_concs": words_error_concs,
        "chunks_raw_text": chunks_raw_text,
    },
    # d. Formatters
    "form": {
        "dummy": lambda x: x,
        "plaintext": plaintext,
        "plaintext_sent_par": plaintext_sent_par,
        "basic_concs": basic_conc,
        "stats_types": stats_types,
    },
}


class Text:
    """
    Takes as input:
        - a string to process
        - the Path object of a file to process

    including a custom pipeline is as simple as:
        - subclassing Text class
        - creating a new @property method like the built in ones while providing your own arguments to self.__process()
    """

    def __init__(self, input, out_file=None, tok_params=None):
        """
        if input == str: return a string
        if input == Path:
                            1. out_file != None: write to given Path object
                            2. out_file == None: write to cwd and append "_pybo" to file name

        custom_tok: settings for building the custom tokenizer: see docstring of Config class in config.py
        """
        self.input = input
        self.tok_params = tok_params

        if isinstance(input, str):
            if out_file:
                assert isinstance(out_file, Path)
                self.out_file = out_file
            else:
                self.out_file = None
        elif isinstance(input, Path):
            if not out_file:
                self.out_file = input.parent / f"{input.stem}_pybo{input.suffix}"
            else:
                self.out_file = out_file
        else:
            raise TypeError("input should either be a string, or a Path object")

    @property
    def tokenize_on_spaces(self):
        return self.__process("basic_cleanup", "space_tok", "dummy", "plaintext")

    @property
    def tokenize_words_raw_text(self):
        return self.__process(
            "basic_cleanup", "word_tok", "words_raw_text", "plaintext",
        )

    @property
    def tokenize_words_raw_lines(self):
        return self.__process(
            "basic_keeps_lines", "word_tok", "words_raw_text", "plaintext",
        )

    @property
    def tokenize_chunks_plaintext(self):
        return self.__process(
            "basic_keeps_lines", "chunk_tok", "chunks_raw_text", "plaintext"
        )

    @property
    def tokenize_sentences_plaintext(self):
        return self.__process(
            "basic_cleanup", "sentence_tok", "dummy", "plaintext_sent_par",
        )

    @property
    def tokenize_paragraph_plaintext(self):
        return self.__process(
            "basic_cleanup", "paragraph_tok", "dummy", "plaintext_sent_par",
        )

    @property
    def list_word_types(self):
        return self.__process(
            "basic_keeps_lines", "word_tok", "words_raw_types", "stats_types",
        )

    def custom_pipeline(
        self, preprocessor, tokenizer, modifier, formatter, tok_params=None
    ):
        """
        every pipe should be either the name of an existing pipe as found in builtin_pipes or a function
        """
        return self.__process(preprocessor, tokenizer, modifier, formatter, tok_params)

    def __process(self, preprocessor, tokenizer, modifier, formatter, tok_params=None):
        if tok_params:
            for k, v in tok_params.items():
                if k not in self.tok_params or self.tok_params[k] is None:
                    self.tok_params[k] = v

        profile, pipes = self.__create_pipeline(
            preprocessor, tokenizer, modifier, formatter, self.tok_params
        )
        pipeline = PipelineBase(profile, pipes=pipes)

        if self.out_file:
            return pipeline.pipe_file(self.input, self.out_file)
        else:
            return pipeline.pipe_str(self.input)

    @staticmethod
    def __create_pipeline(
        preprocessor, tokenizer, modifier, formatter, tok_params=None
    ):
        profile = {}
        pipes = {"prep": {}, "tok": {}, "mod": {}, "form": {}}
        for a, b, c in [
            ("prep", Ids.prep, preprocessor),
            ("tok", Ids.tok, tokenizer),
            ("mod", Ids.mod, modifier),
            ("form", Ids.form, formatter),
        ]:
            if isinstance(c, FunctionType):
                pipes[a].update({b: c})
                profile[a] = b
            elif isinstance(c, str):
                profile[a] = c
                assert c in builtin_pipes[a]
                pipes[a][c] = builtin_pipes[a][c]
            else:
                raise SyntaxError("Should be either a function or a string")

        if tok_params:
            profile["tok_params"] = tok_params
        return profile, pipes



# coding: utf-8
from ..vars import TSEK, AA


class Token:
    def __init__(self):
        self.text = ""
        self.char_types = []
        self.has_merged_dagdra = None
        self.lemma = ""
        self.sense = ""
        self.chunk_type = None
        self.start = 0
        self.len = None
        self.syls_idx = None
        self.syls_start_end = None
        self.pos = ""
        self.affixation = {}
        self.senses = None
        self.affix = False
        self.affix_host = False
        self.form_freq = None
        self.freq = None
        self.skrt = False
        self._ = {}  # dict for any user specific data

    def __getitem__(self, attr):
        # allows to access attributes with the Token['attr'] syntax, besides the Token.attr default
        try:
            return self.__getattribute__(attr)
        except AttributeError:
            raise AttributeError("does not have attribute: " + attr)

    def __setitem__(self, key, value):
        # enforces not to add any extra attribute. Token._ should be used for any custom data
        if hasattr(self, key):
            if key != "_":
                self.__dict__[key] = value
            else:
                if not isinstance(value, dict):
                    raise TypeError("only dicts are accepted for Token._")
                self.__dict__[key].update(value)
        else:
            raise AttributeError("Token objects don't have " + key + " as attribute")

    @property
    def syls(self):
        return (
            [[self.text[s] for s in syl] for syl in self.syls_idx]
            if self.syls_idx
            else ""
        )

    @property
    def text_cleaned(self):
        """
        Will append a TSEK to every syllable except syllables that host
        an affix.

        """
        if self.syls:
            cleaned = TSEK.join(["".join(syl) for syl in self.syls])
            if self.affix_host and not self.affix:
                return cleaned
            else:
                return cleaned + TSEK
        else:
            return ""

    @property
    def text_unaffixed(self):
        unaffixed = TSEK.join(["".join(syl) for syl in self.syls]) if self.syls else ""
        if (
            self.affixation
            and not self.affix
            and "len" in self.affixation
            and len([True for m in self.senses if "affixed" in m and m["affixed"]]) > 0
        ):
            unaffixed = unaffixed[: -self.affixation["len"]]

            if unaffixed and "aa" in self.affixation and self.affixation["aa"]:
                unaffixed += AA

        if self.affixation and self.affix_host and not self.affix:
            return unaffixed
        elif unaffixed:
            return unaffixed + TSEK
        else:
            return ""

    def __repr__(self):
        out = 'text: "{}"\n'.format(self.text)
        if self.text_cleaned:
            out += 'text_cleaned: "{}"\n'.format(self.text_cleaned)
        if self.text_unaffixed:
            out += 'text_unaffixed: "{}"\n'.format(self.text_unaffixed)
        if self.syls and self.syls != []:
            out += (
                'syls: ["' + '", "'.join(["".join(syl) for syl in self.syls]) + '"]\n'
            )
        if self.pos:
            out += "pos: {}\n".format(self.pos)
        if self.lemma:
            out += "lemma: {}\n".format(self.lemma)
        if self.sense:
            out += "sense: {}\n".format(self.sense)
        if self.senses:
            out += (
                "senses: | "
                + " | ".join(
                    [
                        ", ".join([f"{k}: {v}" for k, v in m.items()])
                        for m in self.senses
                    ]
                )
                + " |\n"
            )
        if self.char_types:
            out += "char_types: |" + "|".join(self.char_types) + "|\n"
        if self.chunk_type:
            out += "chunk_type: {}\n".format(self.chunk_type)
        if self.form_freq:
            out += "form_freq: {}\n".format(self.form_freq)
        if self.freq:
            out += "freq: {}\n".format(self.freq)
        if self.skrt:
            out += "skrt: {}\n".format(self.skrt)
        if self.affix:
            out += "affix: {}\n".format(self.affix)
        if self.affix_host:
            out += "affix_host: {}\n".format(self.affix_host)
        if self.has_merged_dagdra:
            out += "has_merged_dagdra: {}\n".format(self.has_merged_dagdra)
        if self.syls_idx:
            out += "syls_idx: {}\n".format(self.syls_idx)
        if self.syls_start_end:
            out += "syls_start_end: {}\n".format(self.syls_start_end)
        out += "start: {}\n".format(self.start)
        out += "len: {}\n".format(self.len)
        if self._:
            out += "\n"
            for k, v in self._.items():
                out += "_{}: {}\n".format(k, v)
        out += "\n"
        return out



import shutil
from pathlib import Path

from ..config import Config


def expose_data(out_path, profile=None):
    """ Copies all the trie and adjustment data to out_path
    :param out_path: must be an existing empty folder
    """
    if profile not in ["POS", "empty"]:
        raise SyntaxError('profile should be either one of ["POS", "empty"]')

    out_path = Path(out_path)
    if not out_path.is_dir() or list(out_path.glob("*")):
        raise IOError("out_path should be an empty folder")

    resources = Path(__file__).parent / "../resources"
    resources = resources.resolve()
    res_dirs = [r for r in resources.glob("*") if r.is_dir()]

    if profile:
        # export profile data
        for f in Config().config["tokenizers"]["profiles"][profile]:
            Path(out_path / Path(f).parent).mkdir(
                parents=True, exist_ok=True
            )  # create dir
            shutil.copy(resources / f, out_path / f)

        shutil.copytree(resources / "adjustment", out_path / "adjustment")

    else:
        # export all data
        for r in res_dirs:
            shutil.copytree(r, out_path / r.name)



#! /usr/bin/env python
# coding: utf8

from __future__ import print_function

import re
from pathlib import Path

import setuptools
from pkg_resources import parse_version

assert parse_version(setuptools.__version__) >= parse_version("38.6.0")


def get_version(prop, project):
    project = Path(__file__).parent / project / "vars.py"
    result = re.search(
        r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format(prop),
        project.read_text(encoding="utf-8-sig"),
    )
    return result.group(1)


def read(fname):
    p = Path(__file__).parent / fname
    with p.open(encoding="utf-8-sig") as f:
        return f.read()


setuptools.setup(
    name="botok",
    version=get_version("__version__", "botok"),  # edit version in botok/vars.py
    author="Esukhia development team",
    author_email="esukhiadev@gmail.com",
    description="Tibetan Word Tokenizer",
    license="Apache2",
    keywords="nlp computational_linguistics tibetan tokenizer token",
    url="https://github.com/Esukhia/botok",
    packages=setuptools.find_packages(),
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    project_urls={
        "Source": "https://github.com/Esukhia/botok",
        "Tracker": "https://github.com/Esukhia/botok/issues",
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Text Processing :: Linguistic",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: Tibetan",
    ],
    package_data={
        "botok": [
            "resources/*",
            "resources/words_bo/*",
            "resources/entry_data/*",
            "resources/words_non_inflected/*",
            "resources/lemmas/*",
            "resources/rules/*",
            "resources/words_skrt/*",
            "resources/adjustment/*",
        ]
    },
    python_requires=">=3.6",
    tests_require=["pytest>=5.0.0"],
    install_requires=["pyyaml", "requests"],
)



# coding: utf-8
from collections import defaultdict
from typing import DefaultDict, List, NewType, Tuple

from ..third_party.has_skrt_syl import has_skrt_syl
from .tokenize import BoToken


def is_mistake(token):
    exceptions = ["\n"]
    if token.chunk_type == "TEXT" or (
        token.chunk_type == "LATIN"
        or token.chunk_type == "CJK"
        or token.chunk_type == "OTHER"
    ):
        if (
            (not token.skrt and not has_skrt_syl(token.text_cleaned))
            and (
                token.senses
                and len(
                    [
                        True
                        for m in token.senses
                        if "pos" in m
                        and (m["pos"] == "NOPOS" or m["pos"] == "NON_WORD")
                    ]
                )
                > 0
                or (
                    token.chunk_type == "LATIN"
                    or token.chunk_type == "CJK"
                    or token.chunk_type == "OTHER"
                )
            )
            and token.text not in exceptions
        ):
            return True
    return False


def words_error_concs(
    tokens: List[BoToken], left=5, right=5
) -> DefaultDict[str, List[str]]:
    mistakes = defaultdict(list)
    for num, t in enumerate(tokens):
        if is_mistake(t):
            if num - left < 0:
                l = tokens[:num]
            else:
                l = tokens[num - left : num]
            if num + right > len(tokens) - 1:
                r = tokens[num + 1 :]
            else:
                r = tokens[num + 1 : num + 1 + right]

            l_context = [t.text for t in l]
            r_context = [t.text for t in r]
            mis = t.text.replace("\n", "\\n")
            mistakes[mis].append(["".join(l_context), "".join(r_context)])
    return mistakes


def words_error_types(tokens: List[BoToken]) -> DefaultDict[str, int]:
    mistakes = defaultdict(int)
    for num, t in enumerate(tokens):
        if is_mistake(t):
            mis = t.text.replace("\n", "\\n")
            mistakes[mis] += 1
    return mistakes


def words_raw_types(tokens: List[BoToken]) -> DefaultDict[str, int]:
    types = defaultdict(int)
    for t in tokens:
        occ = t.text.replace("\n", "\\n")
        types[occ] += 1
    return types


def words_raw_text(tokens: List[BoToken]) -> List[str]:
    return [t.text for t in tokens]


def chunks_raw_text(tokens: List[Tuple[str, str]]) -> List[str]:
    return [chunk for _, chunk in tokens]



# coding: utf-8
from .pynpl.cql import Query


__all__ = ["Query", "parse_cql_query", "replace_token_attributes"]


def parse_cql_query(query, numerals=True, booleans=True):
    """
    CQL parser for replacing the content of Token.attributes.
    From the CQL syntax, only the attribute names and the values
    are taken into account.

    :param query: CQL query string
    :param numerals: if True, gives the Python int instead of the string
    :param booleans: if True, gives the Python bool instead of the string
    :return: a list of dicts, one per token slot, where
                keys == Token.attributes and
                values == content of the expected Token.attributes
    """

    def str2int(string):
        try:
            return int(string)
        except ValueError:
            return string

    def str2bool(string):
        b = {"True": True, "False": False, "None": None}
        if string in b:
            return b[string]
        else:
            return string

    def cql2pattern(tokenexpr, numerals, booleans):
        """
        Expects the following syntax:
            '[attribute1="value1" & attribute2="value2" (& ...)]'
        """
        changes = {}
        for attrexprs in tokenexpr:
            key = attrexprs.attribute
            value = attrexprs.valueexpr[0]
            if numerals:
                value = str2int(value)
            if booleans:
                value = str2bool(value)
            changes[key] = value
        return changes

    if query:
        parsed = Query(query)
        pattern = []
        for tokenexpr in parsed.tokenexprs:
            pattern.append(cql2pattern(tokenexpr, numerals, booleans))
        return pattern
    else:
        return None


def replace_token_attributes(tokens, token_changes):
    """
    Applies in place the replacements found in the CQL query (token_changes)
    the number of tokens in the list and the number of token slots in the query
    must be even.

    :param tokens: list of tokens
    :param token_changes: CQL query
    """
    changes = parse_cql_query(token_changes)
    if type(tokens) == list:
        assert len(tokens) == len(changes)
        for i in range(len(tokens)):
            for attr, value in changes[i].items():
                setattr(tokens[i], attr, value)
    else:
        assert len(changes) == 1
        for attr, value in changes[0].items():
            setattr(tokens, attr, value)



# coding: utf-8
import csv
import pickle
import time
from pathlib import Path

from ..chunks.chunks import TokChunks
from ..vars import HASH, NAMCHE, NO_POS, TSEK, __version__
from .basictrie import BasicTrie, Node


class Trie(BasicTrie):
    def __init__(
        self, bosyl, profile, main_data, custom_data, build=False, pickle_path=None
    ):
        BasicTrie.__init__(self)
        self.bosyl = bosyl()
        self.main_data = main_data
        self.custom_data = custom_data
        self.pickled_file = Path(profile + "_trie.pickled")
        if pickle_path:
            self.pickled_file = Path(pickle_path) / self.pickled_file
        self.tmp_inflected = (
            dict()
        )  # tmp to inflect only once, even if a word appears in many files.
        self.load_or_build_trie(build)

    def rebuild_trie(self):
        self.head = Node()
        self.load_or_build_trie(build=True)

    def load_or_build_trie(self, build=False):
        if build or not self.pickled_file.exists():
            self._build_trie()
        else:
            self._load_trie()

        # add and deactivate the custom entries in memory (will not be written)
        self._populate_trie(self.custom_data)
        self.tmp_inflected = dict()

    def _load_trie(self):
        print("Loading Trie...", end=" ", flush=True)
        start = time.time()
        with self.pickled_file.open("rb") as f:
            self.head = pickle.load(f)
            version = self.head.data["_"]["version"]
            if version != __version__:
                print(
                    f"\nThe trie was build for botok {version}. Current version: {__version__}"
                )
                self._build_trie()
        end = time.time()
        print("({:.0f}s.)".format(end - start), flush=True)

    def _build_trie(self):
        """
        """
        print("Building Trie:", flush=True)
        start = time.time()
        self.head.data["_"]["version"] = __version__  # add version in trie
        self._populate_trie(self.main_data)

        with self.pickled_file.open("wb") as f:
            pickle.dump(self.head, f, pickle.HIGHEST_PROTOCOL)
        end = time.time()
        print("({:.0f} s.)".format(end - start), flush=True)

    def _populate_trie(self, files):
        # first populate the trie with words
        lexica = (d for d in files if d.startswith("lexica"))
        for l in lexica:
            for f in files[l]:
                self._add_one_file(f, l)

        # then add data to the added words
        rest = (
            d for d in files if not d.startswith("lexica") and not d.startswith("rules")
        )
        for r in rest:
            for f in files[r]:
                self._add_one_file(f, r)

    def _add_one_file(self, in_file, category):
        """
        files can have comments starting with #
        spaces and empty lines are trimmed
        a single space(breaks if more than one), a comma or a tab can be used as separators
        """
        print("\t" + str(in_file))
        with in_file.open("r", encoding="utf-8-sig") as f:
            lines = self.__clean_lines(f)
            for l in lines:
                word = l.split("\t", 1)[0]
                if category == "words":
                    self.inflect_n_modify_trie(word)
                    self.inflect_n_add_data(l)

                elif category == "words_non_inflected":
                    self.add_non_inflectible(word)
                    self.inflect_n_add_data(l)

                elif category == "words_skrt":
                    self.inflect_n_modify_trie(word, skrt=True)
                    self.inflect_n_add_data(l)

                elif category == "remove":
                    self.inflect_n_modify_trie(l, deactivate=True)

                else:
                    raise SyntaxError(
                        "'category' is: '"
                        + category
                        + "'. Valid answers: words_bo, words_skrt,"
                        "words_non_inflected, entry_data, remove"
                    )

    def add_non_inflectible(self, word):
        syls = TokChunks(word).get_syls()
        if not syls:
            return None

        infl = self.__join_syls(syls)
        self.add(infl)

    def inflect_n_modify_trie(self, word, deactivate=False, skrt=False):
        """
        Add or deactivate to the trie all the affixed versions of the word
        :param word: a word without ending tsek
        :param deactivate: switch to add or deactivate a word
        """
        inflected = self._get_inflected(word)
        if not inflected:
            return

        for infl, data in inflected:
            if deactivate:
                self.deactivate(infl)
            else:
                if skrt:
                    if data is None:
                        data = {"skrt": True}
                    else:
                        data.update({"skrt": True})
                    self.add(infl, data=data)
                else:
                    self.add(infl, data=data)

    def inflect_n_add_data(self, line):
        form, pos, lemma, sense, freq = self.__parse_line(line)
        freq = int(freq) if freq else None
        lemma = self.__join_syls(TokChunks(lemma).get_syls()) if lemma else None

        inflected = self._get_inflected(form)
        if not inflected:
            return  # The entry is not Tibetan, so return doing nothing

        for infl, _ in inflected:
            affixed = True if _ else False
            data = {
                k: v
                for k, v in [
                    ("lemma", lemma),
                    ("pos", pos),
                    ("freq", freq),
                    ("sense", sense),
                    ("affixed", affixed),
                ]
                if v is not None
            }
            self.add_data(infl, data)

    def _get_inflected(self, word):
        """
        gets the clean syls using TokChunks(), then inflects the last syl using BoSyl.get_all_affixed()

        :return: list of (<inflected word>, <affixation data>)
        """
        if word in self.tmp_inflected:
            return self.tmp_inflected[word]

        syls = TokChunks(word).get_syls()
        if not syls:
            return None

        inflected = [(syls, None)]
        affixed = self.bosyl.get_all_affixed(syls[-1])
        if affixed:
            for infl, data in affixed:
                infl_word = syls[:-1] + [infl]
                inflected.append((infl_word, {"affixation": data}))

        self.tmp_inflected[word] = inflected
        return inflected

    @staticmethod
    def __join_syls(syls):
        return "".join([syl if syl.endswith(NAMCHE) else syl + TSEK for syl in syls])

    @staticmethod
    def __clean_lines(f):
        # cuts off comments, then strips empty lines
        lines = (
            line[: line.index(HASH)] if HASH in line else line for line in f.readlines()
        )
        return (l for l in lines if l)

    @staticmethod
    def __parse_line(line):
        """
        enables support of '\t' and ',' as separator.
        """
        fields = [None, None, None, None, None]
        if "\t" in line:
            sep = "\t"
        elif "," in line:
            sep = ","
        else:
            fields[0] = line
            fields[2] = NO_POS
            return fields

        for num, cell in enumerate(list(csv.reader([line], delimiter=sep))[0]):
            fields[num] = cell if cell else None
        return fields



# coding: utf-8
from typing import DefaultDict, List, Tuple

from .tokenize import BoToken


def plaintext(tokens: List[str], sep=" ") -> str:
    tokens = [t.replace(" ", "_") for t in tokens]
    return sep.join(tokens)


def plaintext_sent_par(units: List[Tuple[int, List[BoToken]]], sep="\n") -> str:
    out = []
    for u in units:
        unit = "".join([word.text for word in u[1]])
        out.append(unit)
    return sep.join(out)


def basic_conc(concs: DefaultDict[str, List[str]], sep="\t", esc_context=True) -> str:
    out = []
    for occ, LR in concs.items():
        for left, right in LR:

            if esc_context:
                left, right = f'"{left}"', f'"{right}"'

            line = f"{left}{sep}{occ}{sep}{right}"
            out.append(line)

    return "\n".join(out)


def stats_types(total_mistakes: DefaultDict[str, int], sep="\t") -> str:
    total = [(mis, freq) for mis, freq in total_mistakes.items()]
    total = sorted(total, reverse=True, key=lambda x: x[1])
    total = [f"{mis}{sep}{freq}" for mis, freq in total]
    return "\n".join(total)



# coding: utf8
from ..chunks.chunks import TokChunks


class ChunkTokenizer(TokChunks):
    def __init__(self, string):
        super().__init__(string)

    def tokenize(self):
        tokens = self.make_chunks()
        return self.get_readable(tokens)



# coding: utf-8
import re

import json
from pathlib import Path

data_path = Path(__file__).parent.parent / "resources" / "SylComponents.json"
with data_path.open("r", encoding="utf-8-sig") as f:
    data = json.loads(f.read())


class SylComponents:
    """
    Provides information about a syllable
    """

    def __init__(self):
        # check for possible dadrag https://github.com/eroux/tibetan-spellchecker/blob/master/doc/second-suffix-da.md
        # roots is an import from root + rareC and wazurC and suffixes is the 'AB' entry from  suffixes.json
        self.dadrag = data["dadrag"]
        self.roots = data["roots"]
        self.suffixes = data["suffixes"]
        self.Csuffixes = data["Csuffixes"]

        # all dicts from https://github.com/eroux/tibetan-spellchecker/tree/master/syllables
        self.special = data["special"]
        self.wazurs = data["wazurs"]
        self.exceptions = self.special + self.wazurs
        self.ambiguous = data["ambiguous"]

        self.m_roots = data["m_roots"]
        self.m_exceptions = data["m_exceptions"]
        self.m_wazurs = data["m_wazurs"]
        self.mingzhis = self.m_roots.copy()
        self.mingzhis.update(self.m_exceptions)
        self.mingzhis.update(self.m_wazurs)

    def get_parts(self, syl):
        """
        :param syl: takes a syllable as input
        :return: a tuple:
                        - (prefix+main-stack, vowel+suffixes)
                        - (exceptions, x)
                        - a list of solutions if there is more than one
                        - None if the syllable is not wellformed
        """
        if syl not in self.exceptions and syl not in self.ambiguous:
            l_s = len(syl)
            # find all possible roots
            root = []
            if len(syl) > 5 and syl[:6] in self.roots:
                root.append(syl[:6])
            if len(syl) > 4 and syl[:5] in self.roots:
                root.append(syl[:5])
            if len(syl) > 3 and syl[:4] in self.roots:
                root.append(syl[:4])
            if len(syl) > 2 and syl[:3] in self.roots:
                root.append(syl[:3])
            if len(syl) > 1 and syl[:2] in self.roots:
                root.append(syl[:2])
            if len(syl) > 0 and syl[:1] in self.roots:
                root.append(syl[:1])
            # find all possible suffixes
            suffix = []
            if l_s > 1:
                if syl[l_s - 1 :] in self.suffixes:
                    suffix.append(syl[l_s - 1 :])
                if syl[l_s - 2 :] in self.suffixes:
                    suffix.append(syl[l_s - 2 :])
                if syl[l_s - 3 :] in self.suffixes:
                    suffix.append(syl[l_s - 3 :])
                if syl[l_s - 4 :] in self.suffixes:
                    suffix.append(syl[l_s - 4 :])
                if syl[l_s - 5 :] in self.suffixes:
                    suffix.append(syl[l_s - 5 :])

            # deal with all the C roots
            # print(self.syl, root)
            if root != [] and self.roots[root[0]] == "C":
                if root[0] == syl:
                    return root[0], ""
                else:
                    for s in suffix:
                        if s in self.Csuffixes and root[0] + s == syl:
                            return root[0], s

            # find all possible matches
            solutions = []
            if suffix != [] and root != []:
                # dealing with all other cases
                for r in root:
                    for s in suffix:
                        # unexpected འ་
                        if self.roots[r] == "A" and s == "འ" and r + s == syl:
                            # print(r, roots[r])
                            return None
                        else:
                            # if root+suffix make the syllable + avoids duplicates
                            if r + s == syl and (r, s) not in solutions:
                                solutions.append((r, s))
                if solutions != []:
                    if len(solutions) > 1:
                        return solutions
                    else:
                        return solutions[0]
                # root + suffix don’t make syl
                else:
                    # print(solutions)
                    return None
            elif root != []:
                # print('k')
                for r in root:
                    if r in self.roots and r == syl:
                        # if syllable is valid without suffix + without aa
                        if self.roots[r] != "NB" and (r, "") not in solutions:
                            solutions.append((r, ""))
                if solutions != []:
                    if len(solutions) > 1:
                        return solutions
                    else:
                        return solutions[0]
                # non-valid syl
                else:
                    return None
            # non-valid syl
            else:
                return None
        elif syl in self.ambiguous:
            return self.ambiguous[syl]
        else:
            return syl, "x"

    def normalize_dadrag(self, syl):
        #
        components = self.get_parts(syl)
        if type(components) == list or not components:
            if syl[-1] == "ད":
                components = self.get_parts(syl[:-1])
                if components and isinstance(components, tuple):
                    return syl[:-1]
        return syl

    def get_mingzhi(self, syl):
        """
        :param syl: syllable
        :return:    the mingzhi that will serve for the particle agreement. for example, ཁྱེའུར will return འ
                    None if more than one solution from get_parts()
        """
        components = self.get_parts(syl)
        if isinstance(components, list) or not components:
            if syl and syl[-1] == "ད":
                components = self.get_parts(syl[:-1])
                if isinstance(components, list) or not components:
                    return None
                else:
                    return self.mingzhis[components[0]]
            return None
        else:
            return self.mingzhis[components[0]]

    def get_info(self, syl):
        """
        :param syl: syllable
        :return: required info to part_agreement:
                - "dadrag" A syllable that historically received a "da" second suffix.
                            As for now, the list contains ["ཀུན", "ཤིན", "འོན"] (See pybo/resources/SylComponents.json)
                - "thame"  A syllable that has the potential of hosting an affixed particle.
                            Will be returned for all such syls, whether or not there is a particle.
                - the syllable itself in all other cases
        """
        mingzhi = self.get_mingzhi(syl)
        if not mingzhi:
            return None
        else:
            if syl in self.dadrag:
                return "dadrag"
            elif (
                re.search(
                    mingzhi
                    + "([ྱྲླྭྷ]?[ིེོུ]?(འ?[ིོུ]?ར?ས?|(འ[མང])|(འོའ[མང])|(འིའ[ོམང])))$",
                    syl,
                )
                is not None
            ):
                return "thame"
            else:
                return syl

    def is_thame(self, syl):
        """
        :param syl: a string without tsek or other punct
        :return: True if the syllabe is affixable or is already affixed, False otherwise
        """
        return self.get_info(syl) == "thame"


if __name__ == "__main__":
    # example of use
    sc = SylComponents()
    assert sc.get_parts("བཀྲིས") == ("བཀྲ", "ིས")
    assert sc.get_info("དེའིའམ") == "thame"
    assert sc.get_info("དེའི") == "thame"



# coding: utf-8
import re

from ..vars import TSEK


def is_skrt(syl):
    """Source for regexes : Paul Hackett Visual Basic script

    regex1: Now do Sanskrit: Skt.vowels, [g|d|b|dz]+_h, hr, shr, Skt
    regex2: more Sanskrit: invalid superscript-subscript pairs
    regex3: tsa-phru mark used in Chinese transliteration
    :param syl: syllable to assert
    :return: True if matches either of the regexes, False otherwise
    """
    regex1 = (
        r"([ཀ-ཬཱ-྅ྐ-ྼ]{0,}[ཱཱཱིུ-ཹཻཽ-ྃ][ཀ-ཬཱ-྅ྐ-ྼ]{0,}|[ཀ-ཬཱ-྅ྐ-ྼ]{0,}"
        r"[གཌདབཛྒྜྡྦྫ][ྷ][ཀ-ཬཱ-྅ྐ-ྼ]{0,}|[ཀ-ཬཱ-྅ྐ-ྼ]{0,}[ཤཧ][ྲ][ཀ-ཬཱ-྅ྐ-ྼ]{0,}|[ཀ-ཬཱ-྅ྐ-ྼ]{0,}"
        r"[གྷཊ-ཎདྷབྷཛྷཥཀྵ-ཬཱཱཱིུ-ཹཻཽ-ྃྒྷྚ-ྞྡྷྦྷྫྷྵྐྵ-ྼ][ཀ-ཬཱ-྅ྐ-ྼ]{0,})"
    )
    regex2 = r"([ཀ-ཬཱ-྅ྐ-ྼ]{0,}[ཀཁགང-ཉཏ-དན-བམ-ཛཝ-ཡཤཧཨ][ྐ-ྫྷྮ-ྰྴ-ྼ][ཀ-ཬཱ-྅ྐ-ྼ]{0,})"
    regex3 = r"([ཀ-ཬཱ-྅ྐ-ྼ]{0,}[༹][ཀ-ཬཱ-྅ྐ-ྼ]{0,})"
    return re.search(regex1, syl) or re.search(regex2, syl) or re.search(regex3, syl)


def has_skrt_syl(word):
    """Uses is_skrt() to check for sanskrit syllables

    """
    skrt = False
    syls = word.strip(TSEK).split(TSEK)
    for s in syls:
        if is_skrt(s):
            skrt = True

    return skrt



# coding: utf-8
from ..third_party.cqlparser import Query
from .tokensplit import TokenSplit


class SplittingMatcher:
    def __init__(self, query, replace_idx, split_idx, token_list, token_changes=None):
        self.matcher = Query(query)
        self.span = len(self.matcher.tokenexprs) - 1
        self.token_list = token_list

        self.replace_idx = replace_idx - 1
        self.split_idx = split_idx
        self.token_changes = token_changes

    def split_on_matches(self, mode="char"):
        """
        :param mode: can either be "char" or "syl"
        """
        split_list = []

        i = 0
        while i < len(self.token_list):
            if self.__matches(i, self.token_list):
                # find the index of the token to split
                idx = i + self.replace_idx

                # add new tokens that precede the one to split
                for r in range(i, idx):
                    split_list.append(self.token_list[r])
                    i += 1

                # split the token and add them to the new list
                split_list.extend(self.__split(self.token_list[idx], mode=mode))

            else:
                split_list.append(self.token_list[i])
            i += 1

        return split_list

    def __matches(self, i, token_list):
        return i + self.span <= len(token_list) and self.matcher(
            self.token_list[i : i + self.span + 1]
        )

    def __split(self, token, mode):
        ts = TokenSplit(token, self.split_idx, self.token_changes)
        return ts.split(mode=mode)



# coding: utf8
from .sentencetokenizer import get_sentence_indices


def paragraph_tokenizer(tokens):
    # a paragraph is defined as a group of sentences that does not have more words than a given threshold
    threshold = 70
    paragraph_max = 150
    par_indices = get_sentence_indices(tokens)

    # join small sentences to form paragraphs
    i = 0
    while i < len(par_indices):
        start, end, l = (
            par_indices[i]["start"],
            par_indices[i]["end"],
            par_indices[i]["len"],
        )
        if i > 0 and l < threshold:
            previous_len = par_indices[i - 1]["len"]
            if l + previous_len < paragraph_max:
                par_indices[i - 1]["end"] = par_indices[i]["end"]
                par_indices[i - 1]["len"] += par_indices[i]["len"]
                del par_indices[i]
                i -= 1
        i += 1

    # get tokens for each paragraph
    pars = []
    for par in par_indices:
        start, end, l = par["start"], par["end"], par["len"]
        pars.append((l, tokens[start : end + 1]))

    return pars
